# Copyright (c) 2019 Sean Vig

from .compositor import Compositor  # noqa: F401
from .cursor import Cursor  # noqa: F401
from .data_control_v1 import DataControlManagerV1  # noqa: F401
from .data_device_manager import DataDeviceManager  # noqa: F401
from .export_dmabuf_v1 import ExportDmabufManagerV1  # noqa: F401
from .foreign_toplevel_management_v1 import ForeignToplevelManagerV1  # noqa: F401
from .gamma_control_v1 import GammaControlManagerV1  # noqa: F401
from .input_device import InputDevice  # noqa: F401
from .input_inhibit import InputInhibitManager  # noqa: F401
from .keyboard import Keyboard  # noqa: F401
from .layer_shell_v1 import LayerShellV1  # noqa: F401
from .matrix import Matrix  # noqa: F401
from .output import Output  # noqa: F401
from .output_damage import OutputDamage  # noqa: F401
from .output_layout import OutputLayout  # noqa: F401
from .pointer import (  # noqa: F401
    PointerEventAxis,
    PointerEventButton,
    PointerEventMotion,
    PointerEventMotionAbsolute,
)
from .pointer_constraints_v1 import (  # noqa: F401
    PointerConstraintsV1,
    PointerConstraintV1,
)
from .primary_selection_v1 import PrimarySelectionV1DeviceManager  # noqa: F401
from .relative_pointer_manager_v1 import RelativePointerManagerV1  # noqa: F401
from .scene import Scene, SceneNode  # noqa: F401
from .screencopy_v1 import ScreencopyManagerV1  # noqa: F401
from .seat import Seat  # noqa: F401
from .surface import Surface, SurfaceState  # noqa: F401
from .texture import Texture  # noqa: F401
from .virtual_keyboard_v1 import VirtualKeyboardManagerV1  # noqa: F401
from .virtual_pointer_v1 import VirtualPointerManagerV1  # noqa: F401
from .xcursor_manager import XCursorManager  # noqa: F401
from .xdg_decoration_v1 import XdgDecorationManagerV1  # noqa: F401
from .xdg_output_v1 import XdgOutputManagerV1  # noqa: F401
from .xdg_shell import XdgShell  # noqa: F401


def __getattr__(name: str):
    if name == "Box":
        from .box import Box  # noqa: F401

        return Box
    try:
        return globals()[name]
    except KeyError:
        raise ImportError(f"cannot import name '{name}' from wlroots.wlr_types")



# Copyright (c) Sean Vig 2020

import enum

from wlroots import lib


@enum.unique
class Edges(enum.IntFlag):
    NONE = lib.WLR_EDGE_NONE
    TOP = lib.WLR_EDGE_TOP
    BOTTOM = lib.WLR_EDGE_BOTTOM
    LEFT = lib.WLR_EDGE_LEFT
    RIGHT = lib.WLR_EDGE_RIGHT



# Copyright (c) 2019 Sean Vig

from __future__ import annotations

from pywayland.server import Display
from typing import TYPE_CHECKING

from wlroots import lib, Ptr

if TYPE_CHECKING:
    from wlroots.renderer import Renderer  # noqa: F401


class Compositor(Ptr):
    def __init__(self, display: Display, renderer: Renderer) -> None:
        """A compositor for clients to be able to allocate surfaces

        :param display:
            The Wayland server display to attach to the compositor.
        :param renderer:
            The wlroots renderer to attach the compositor to.
        """
        self._ptr = lib.wlr_compositor_create(display._ptr, renderer._ptr)



# Copyright (c) Matt Colligan 2021

from __future__ import annotations

from pywayland.protocol.wayland import WlOutput
from wlroots import ffi, lib, Ptr
from wlroots.util.box import Box


class PixmanRegion32(Ptr):
    def __init__(self, ptr=None) -> None:
        """This is a convenience wrapper around pixman_region32_t

        :param ptr:
            The pixman_region32_t cdata pointer
        """
        if ptr is None:
            self._ptr = ffi.new("struct pixman_region32 *")
        else:
            self._ptr = ptr

    def __enter__(self) -> PixmanRegion32:
        """Use the pixman_region32 in a context manager"""
        lib.pixman_region32_init(self._ptr)
        return self

    def __exit__(self, exc_type, exc_value, exc_tb) -> None:
        """Finish up when exiting the context"""
        lib.pixman_region32_fini(self._ptr)

    def rectangles_as_boxes(self) -> list[Box]:
        nrects_ptr = ffi.new("int *")
        rects = lib.pixman_region32_rectangles(self._ptr, nrects_ptr)
        nrects = nrects_ptr[0]

        boxes = []
        for i in range(nrects):
            boxes.append(
                Box(rects.x1, rects.y1, rects.x2 - rects.x1, rects.y2 - rects.y1)
            )
            rects += 1
        return boxes

    def transform(
        self,
        src: PixmanRegion32,
        transform: WlOutput.transform,
        width: int,
        height: int,
    ) -> None:
        """
        Applies a transform to a region inside a box of size `width` x `height`.
        """
        lib.wlr_region_transform(self._ptr, src._ptr, transform, width, height)

    def not_empty(self) -> bool:
        """
        Wrapper around pixman_region32_not_empty
        """
        return lib.pixman_region32_not_empty(self._ptr)



# Copyright Sean Vig (c) 2020
# Copyright Matt Colligan (c) 2021

from __future__ import annotations

from wlroots import ffi, lib


def _int_getter(attr):
    def getter(self):
        return getattr(self._ptr, attr)

    return getter


def _int_setter(attr):
    def setter(self, value):
        setattr(self._ptr, attr, value)

    return setter


class Box:
    def __init__(
        self,
        x: int | None = None,
        y: int | None = None,
        width: int | None = None,
        height: int | None = None,
        ptr=None,
    ) -> None:
        """A simple box structure, represented by a coordinate and dimensions"""
        if ptr is None:
            self._ptr = ffi.new("struct wlr_box *")
        else:
            self._ptr = ptr

        if x is not None:
            self.x = x
        if y is not None:
            self.y = y
        if width is not None:
            self.width = width
        if height is not None:
            self.height = height

    x = property(_int_getter("x"), _int_setter("x"))
    y = property(_int_getter("y"), _int_setter("y"))
    width = property(_int_getter("width"), _int_setter("width"))
    height = property(_int_getter("height"), _int_setter("height"))

    def __repr__(self) -> str:
        return "Box(%s, %s, %s, %s)" % (self.x, self.y, self.width, self.height)

    def closest_point(self, x: float, y: float) -> tuple[float, float]:
        xy_ptr = ffi.new("double[2]")
        lib.wlr_box_closest_point(self._ptr, x, y, xy_ptr, xy_ptr + 1)
        return xy_ptr[0], xy_ptr[1]

    def contains_point(self, x: float, y: float) -> bool:
        return lib.wlr_box_contains_point(self._ptr, x, y)



from __future__ import annotations

from typing import TYPE_CHECKING

from pywayland.server import Listener

if TYPE_CHECKING:
    from wlroots.wlr_types import InputDevice, Keyboard
    from wlroots.wlr_types.keyboard import KeyboardKeyEvent

    from .server import TinywlServer


class KeyboardHandler:
    def __init__(
        self,
        keyboard: Keyboard,
        input_device: InputDevice,
        tinywl_server: TinywlServer,
    ) -> None:
        self.keyboard = keyboard
        self.input_device = input_device
        self.tinywl_server = tinywl_server

        keyboard.modifiers_event.add(Listener(self.keyboard_handle_modifiers))
        keyboard.key_event.add(Listener(self.keyboard_handle_key))

    def keyboard_handle_modifiers(self, listener: Listener, data) -> None:
        """Activates the keyboard and sends the modifiers event to the active surface"""
        self.tinywl_server.send_modifiers(self.keyboard.modifiers, self.input_device)

    def keyboard_handle_key(
        self, listener: Listener, key_event: KeyboardKeyEvent
    ) -> None:
        """Activates the keyboard and sends the key event to the active surface

        Gives the display manager the first right of refusal
        """
        self.tinywl_server.send_key(key_event, self.input_device)



# Copyright Matt Colligan (c) 2021
#
# The contents of this module has been moved to ../util/box.py for wlroots 0.15.0 and
# this file will be removed in the future.

import warnings

from wlroots.util.box import *  # noqa: F401,F403

warnings.warn(
    "wlroots.types.box has moved to wlroots.util.box and will be removed in the future.",
    DeprecationWarning,
    stacklevel=1,
)



# Copyright (c) 2021 Sean Vig

from __future__ import annotations

from pywayland.server import Display

from wlroots.allocator import Allocator
from wlroots.backend import Backend, BackendType
from wlroots.renderer import Renderer
from wlroots.wlr_types import Compositor


def build_compositor(
    display: Display, *, backend_type=BackendType.AUTO
) -> tuple[Compositor, Allocator, Renderer, Backend]:
    """Build and run a compositor

    :param display:
        The Wayland display to attatch to the backend, renderer, and
        compositor.
    :param backend_type:
        The type of the backend to setup the compositor for, by default use the
        auto-detected backend.
    :return:
        The compositor, allocator, renderer, and the backend.
    """
    backend = Backend(display, backend_type=backend_type)
    renderer = Renderer.autocreate(backend)
    renderer.init_display(display)
    allocator = Allocator.autocreate(backend, renderer)
    compositor = Compositor(display, renderer)

    return compositor, allocator, renderer, backend



# Copyright (c) Matt Colligan 2021
#
# This protocol is obsolete and will be removed in a future version. The recommended
# replacement is xdg-decoration.

from __future__ import annotations

import enum
from typing import TYPE_CHECKING

from wlroots import PtrHasData, lib

if TYPE_CHECKING:
    from pywayland.server import Display


class ServerDecorationManagerMode(enum.IntEnum):
    NONE = lib.WLR_SERVER_DECORATION_MANAGER_MODE_NONE
    CLIENT = lib.WLR_SERVER_DECORATION_MANAGER_MODE_CLIENT
    SERVER = lib.WLR_SERVER_DECORATION_MANAGER_MODE_SERVER


class ServerDecorationManager(PtrHasData):
    def __init__(self, ptr) -> None:
        """
        A decoration negotiation interface which implements the KDE protocol:
        wlr_server_decoration_manager.
        """
        self._ptr = ptr

    @classmethod
    def create(cls, display: Display) -> ServerDecorationManager:
        """Create a wlr_server_decoration_manager for the given display."""
        ptr = lib.wlr_server_decoration_manager_create(display._ptr)
        return cls(ptr)

    def set_default_mode(self, default_mode: ServerDecorationManagerMode) -> None:
        """Set the default decoration mode for this server"""
        lib.wlr_server_decoration_manager_set_default_mode(self._ptr, default_mode)



# Copyright (c) 2021 Matt Colligan

from pywayland.server import Display, Signal

from wlroots import ffi, lib, Ptr
from .output_layout import OutputLayout


class XdgOutputManagerV1(Ptr):
    def __init__(self, display: Display, layout: OutputLayout) -> None:
        """Create an wlr_xdg_output_manager_v1"""
        self._ptr = lib.wlr_xdg_output_manager_v1_create(display._ptr, layout._ptr)

        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))



# Copyright (c) Matt Colligan 2021
from __future__ import annotations

from dataclasses import dataclass
from typing import Iterator

from pywayland.server import Display, Signal
from pywayland.protocol.wayland import WlOutput
from pywayland.utils import wl_list_for_each

from wlroots import ffi, lib, Ptr
from .output import Output, OutputMode


@dataclass
class CustomMode:
    """The custom_mode struct member of wlr_output_state"""

    width: int
    height: int
    refresh: int


class OutputHeadV1State(Ptr):
    def __init__(self, ptr) -> None:
        """wlr_output_head_v1_state"""
        self._ptr = ptr

    @property
    def enabled(self) -> bool:
        return self._ptr.enabled

    @enabled.setter
    def enabled(self, value: bool) -> None:
        self._ptr.enabled = value

    @property
    def x(self) -> int:
        return self._ptr.x

    @x.setter
    def x(self, value: int) -> None:
        self._ptr.x = value

    @property
    def y(self) -> int:
        return self._ptr.y

    @y.setter
    def y(self, value: int) -> None:
        self._ptr.y = value

    @property
    def scale(self) -> float:
        return self._ptr.scale

    @scale.setter
    def scale(self, value: float) -> None:
        self._ptr.scale = value

    @property
    def output(self) -> Output:
        return Output(self._ptr.output)

    @property
    def mode(self) -> OutputMode:
        return OutputMode(self._ptr.mode)

    @mode.setter
    def mode(self, value: OutputMode) -> None:
        self._ptr.mode = value._ptr

    @property
    def custom_mode(self):
        width = self._ptr.custom_mode.width
        height = self._ptr.custom_mode.height
        refresh = self._ptr.custom_mode.refresh
        return CustomMode(width, height, refresh)

    @custom_mode.setter
    def custom_mode(self):
        width = self._ptr.custom_mode.width
        height = self._ptr.custom_mode.height
        refresh = self._ptr.custom_mode.refresh
        return CustomMode(width, height, refresh)

    @property
    def transform(self) -> WlOutput.transform:
        return WlOutput.transform(self._ptr.transform)

    @transform.setter
    def transform(self, value: WlOutput.transform) -> None:
        self._ptr.transform = value


class OutputConfigurationV1(Ptr):
    def __init__(self, ptr=None) -> None:
        """wlr_output_configuration_v1

        If a pointer is not given, a new instance is created. Pointers are given when
        these are returned by OutputManagerV1's events.
        """
        if ptr is None:
            self._ptr = lib.wlr_output_configuration_v1_create()
        else:
            self._ptr = ffi.cast("struct wlr_output_configuration_v1 *", ptr)

    @property
    def heads(self) -> Iterator["OutputConfigurationHeadV1"]:
        for ptr in wl_list_for_each(
            "struct wlr_output_configuration_head_v1 *",
            self._ptr.heads,
            "link",
            ffi=ffi,
        ):
            yield OutputConfigurationHeadV1(ptr)

    def send_succeeded(self) -> None:
        """
        If the configuration comes from a client request, this sends positive
        feedback to the client (configuration has been applied).
        """
        lib.wlr_output_configuration_v1_send_succeeded(self._ptr)

    def send_failed(self) -> None:
        """
        If the configuration comes from a client request, this sends negative
        feedback to the client (configuration has not been applied).
        """
        lib.wlr_output_configuration_v1_send_failed(self._ptr)

    def destroy(self) -> None:
        """Destroy the instance."""
        lib.wlr_output_configuration_v1_destroy(self._ptr)


class OutputConfigurationHeadV1(Ptr):
    def __init__(self, ptr) -> None:
        """An instance of wlr_output_configuration_head_v1"""
        self._ptr = ptr

    @classmethod
    def create(
        cls, config: OutputConfigurationV1, output: Output
    ) -> OutputConfigurationHeadV1:
        """Create a new wlr_output_configuration_head_v1"""
        ptr = lib.wlr_output_configuration_head_v1_create(config._ptr, output._ptr)
        return OutputConfigurationHeadV1(ptr)

    @property
    def state(self) -> OutputHeadV1State:
        return OutputHeadV1State(self._ptr.state)


class OutputManagerV1(Ptr):
    def __init__(self, display: Display) -> None:
        """Create a wlr_output_manager_v1

        Create a new output manager. The compositor is responsible for calling
        `wlr_output_manager_v1_set_configuration` whenever the current output
        configuration changes.
        """
        self._ptr = lib.wlr_output_manager_v1_create(display._ptr)

        self.apply_event = Signal(
            ptr=ffi.addressof(self._ptr.events.apply),
            data_wrapper=OutputConfigurationV1,
        )
        self.test_event = Signal(
            ptr=ffi.addressof(self._ptr.events.test), data_wrapper=OutputConfigurationV1
        )

    def set_configuration(self, config: OutputConfigurationV1) -> None:
        """
        Updates the output manager's current configuration. This will broadcast any
        changes to all clients.

        This function takes ownership over `config`. That is, the compositor must not
        access the configuration anymore.
        """
        lib.wlr_output_manager_v1_set_configuration(self._ptr, config._ptr)



# Copyright Sean Vig (c) 2020

from __future__ import annotations

from pywayland.protocol.wayland import WlOutput

from wlroots import ffi, lib, Ptr
from wlroots.util.box import Box


class Matrix(Ptr):
    def __init__(self, ptr) -> None:
        """A matrix which encodes transformations used for rendering"""
        self._ptr = ptr

    @classmethod
    def identity(cls) -> Matrix:
        """An identity matrix"""
        mat_ptr = cls._build_matrix_ptr()
        lib.wlr_matrix_identity(mat_ptr)
        return Matrix(mat_ptr)

    @classmethod
    def projection(
        cls, width: int, height: int, transform: WlOutput.transform
    ) -> Matrix:
        """A 2d orthographic projection matrix of (width, height) with specified transform"""
        mat_ptr = cls._build_matrix_ptr()
        lib.wlr_matrix_projection(mat_ptr, width, height, transform)
        return Matrix(mat_ptr)

    @classmethod
    def project_box(
        cls,
        box: Box,
        transform: WlOutput.transform,
        rotation: float,
        projection: Matrix,
    ) -> Matrix:
        """Project the specified box onto a orthographic projection with a rotation"""
        mat_ptr = cls._build_matrix_ptr()
        lib.wlr_matrix_project_box(
            mat_ptr, box._ptr, transform, rotation, projection._ptr
        )
        return Matrix(mat_ptr)

    def transpose(self) -> Matrix:
        """Transpose the matrix"""
        mat_ptr = self._build_matrix_ptr()
        lib.wlr_matrix_transpose(mat_ptr, self._ptr)
        return Matrix(mat_ptr)

    def translate(self, x: float, y: float) -> None:
        """A 2d translation matrix of magnitude (x,y)"""
        lib.wlr_matrix_translate(self._ptr, x, y)

    def scale(self, x: float, y: float) -> None:
        """A 2d scale matrix of magnitude (x,y)"""
        lib.wlr_matrix_scale(self._ptr, x, y)

    def rotate(self, rad: float) -> None:
        """A 2d rotation of angle rad"""
        lib.wlr_matrix_rotate(self._ptr, rad)

    def transform(self, transform: WlOutput.transform) -> None:
        """Apply the given transformation to the matrix"""
        lib.wlr_matrix_transform(self._ptr, transform)

    def __matmul__(self, other: Matrix) -> Matrix:
        """Perform matrix multiplication with given matrix"""
        mat_ptr = self._build_matrix_ptr()
        lib.wlr_matrix_multiply(mat_ptr, self._ptr, other._ptr)
        return Matrix(mat_ptr)

    def __str__(self) -> str:
        """String representation showing the matrix"""
        matrix_values = list(self._ptr)
        return (
            f"Matrix([{matrix_values[:3]}, {matrix_values[3:6]}, {matrix_values[6:]}])"
        )

    @staticmethod
    def _build_matrix_ptr():
        return ffi.new("float [9]")



# Copyright (c) Sean Vig 2019

from __future__ import annotations

from typing import TYPE_CHECKING

from wlroots import ffi, lib, Ptr

from .cursor import Cursor

if TYPE_CHECKING:
    from typing import Iterator


class XCursorManager(Ptr):
    def __init__(self, size, scale=1):
        """Creates a new XCursor manager

        Create cursor with base size and scale.
        """
        ptr = lib.wlr_xcursor_manager_create(ffi.NULL, size)
        self._ptr = ffi.gc(ptr, lib.wlr_xcursor_manager_destroy)

        lib.wlr_xcursor_manager_load(self._ptr, scale)

    def destroy(self):
        """Destroy the x cursor manager"""
        if self._ptr is not None:
            ffi.release(self._ptr)
            self._ptr = None

    def set_cursor_image(self, name: str, cursor: Cursor):
        """Set the cursor image

        Set a Cursor image to the specified cursor name for all scale factors.
        The wlroots cursor will take over from this point and ensure the
        correct cursor is used on each output, assuming an output layout is
        attached to it.
        """
        name_cdata = ffi.new("char []", name.encode())
        lib.wlr_xcursor_manager_set_cursor_image(self._ptr, name_cdata, cursor._ptr)

    def get_xcursor(self, name: str, scale: float = 1) -> XCursor | None:
        """
        Retrieves a wlr_xcursor reference for the given cursor name at the given scale
        factor, or NULL if this wlr_xcursor_manager has not loaded a cursor theme at the
        requested scale.
        """
        ptr = lib.wlr_xcursor_manager_get_xcursor(self._ptr, name.encode(), scale)
        if ptr == ffi.NULL:
            return None
        return XCursor(ptr)

    def __enter__(self):
        """Setup X cursor manager in a context manager"""
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """Clean-up the X cursor manager on contex manager exit"""
        self.destroy()


class XCursor(Ptr):
    """struct wlr_xcursor"""

    def __init__(self, ptr):
        self._ptr = ptr

    @property
    def images(self) -> Iterator[XCursorImage]:
        for i in range(self._ptr.image_count):
            yield XCursorImage(self._ptr.images[0][i])


class XCursorImage(Ptr):
    """struct wlr_xcursor_image"""

    def __init__(self, ptr):
        self._ptr = ptr



# Copyright (c) Sean Vig 2019

from __future__ import annotations

import enum
import weakref
from typing import Callable, TypeVar

from pywayland.server import Display, Signal

from wlroots import ffi, PtrHasData, lib, Ptr, str_or_none
from wlroots.util.box import Box
from wlroots.util.edges import Edges
from .output import Output
from .surface import Surface

_weakkeydict: weakref.WeakKeyDictionary = weakref.WeakKeyDictionary()

T = TypeVar("T")
SurfaceCallback = Callable[[Surface, int, int, T], None]


@ffi.def_extern()
def surface_iterator_callback(surface_ptr, sx, sy, data_ptr):
    """Callback used to invoke the for_each_surface method"""
    func, py_data = ffi.from_handle(data_ptr)
    surface = Surface(surface_ptr)
    func(surface, sx, sy, py_data)


class XdgSurfaceRole(enum.IntEnum):
    NONE = lib.WLR_XDG_SURFACE_ROLE_NONE
    TOPLEVEL = lib.WLR_XDG_SURFACE_ROLE_TOPLEVEL
    POPUP = lib.WLR_XDG_SURFACE_ROLE_POPUP


class XdgShell(PtrHasData):
    def __init__(self, display: Display) -> None:
        """Create the shell for protocol windows

        :param display:
            The Wayland server display to create the shell on.
        """
        self._ptr = lib.wlr_xdg_shell_create(display._ptr)

        self.new_surface_event = Signal(
            ptr=ffi.addressof(self._ptr.events.new_surface), data_wrapper=XdgSurface
        )
        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))


class XdgSurface(PtrHasData):
    def __init__(self, ptr) -> None:
        """A user interface element requiring management by the compositor

        An xdg-surface is a user interface element requiring management by the
        compositor. An xdg-surface alone isn't useful, a role should be
        assigned to it in order to map it.

        When a surface has a role and is ready to be displayed, the `map` event
        is emitted. When a surface should no longer be displayed, the `unmap`
        event is emitted. The `unmap` event is guaranteed to be emitted before
        the `destroy` event if the view is destroyed when mapped
        """
        self._ptr = ffi.cast("struct wlr_xdg_surface *", ptr)

        self.map_event = Signal(ptr=ffi.addressof(self._ptr.events.map))
        self.unmap_event = Signal(ptr=ffi.addressof(self._ptr.events.unmap))
        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))
        self.new_popup_event = Signal(
            ptr=ffi.addressof(self._ptr.events.new_popup), data_wrapper=XdgPopup
        )
        self.configure_event = Signal(
            ptr=ffi.addressof(self._ptr.events.configure),
            data_wrapper=XdgSurfaceConfigure,
        )
        self.ack_configure_event = Signal(
            ptr=ffi.addressof(self._ptr.events.ack_configure),
            data_wrapper=XdgSurfaceConfigure,
        )

    @classmethod
    def from_surface(cls, surface: Surface) -> XdgSurface:
        """Get the xdg surface associated with the given surface"""
        if not surface.is_xdg_surface:
            raise RuntimeError("Surface is not XDG surface")
        surface_ptr = lib.wlr_xdg_surface_from_wlr_surface(surface._ptr)
        return XdgSurface(surface_ptr)

    @property
    def surface(self) -> Surface:
        """The surface associated with the xdg surface"""
        return Surface(self._ptr.surface)

    @property
    def role(self) -> XdgSurfaceRole:
        """The role for the surface"""
        return XdgSurfaceRole(self._ptr.role)

    @property
    def toplevel(self) -> XdgTopLevel:
        """Return the top level xdg object

        This shell must be a top level role
        """
        if self.role != XdgSurfaceRole.TOPLEVEL:
            raise ValueError(f"xdg surface must be top-level, got: {self.role}")

        toplevel = XdgTopLevel(self._ptr.toplevel)

        # the toplevel does not own the ptr data, ensure the underlying cdata
        # is kept alive
        _weakkeydict[toplevel] = self

        return toplevel

    @property
    def popup(self) -> XdgPopup:
        """Return the popup xdg object

        This shell must be a popup role.
        """
        if self.role != XdgSurfaceRole.POPUP:
            raise ValueError(f"xdg surface must be popup, got: {self.role}")

        popup = XdgPopup(self._ptr.popup)
        _weakkeydict[popup] = self

        return popup

    def get_geometry(self) -> Box:
        """Get the surface geometry

        This is either the geometry as set by the client, or defaulted to the
        bounds of the surface + the subsurfaces (as specified by the protocol).

        The x and y value can be <0
        """
        box_ptr = ffi.new("struct wlr_box *")
        lib.wlr_xdg_surface_get_geometry(self._ptr, box_ptr)
        return Box(box_ptr.x, box_ptr.y, box_ptr.width, box_ptr.height)

    def set_size(self, width: int, height: int) -> int:
        return lib.wlr_xdg_toplevel_set_size(self._ptr, width, height)

    def set_activated(self, activated: bool) -> int:
        if self.role != XdgSurfaceRole.TOPLEVEL:
            raise ValueError(f"xdg surface must be top-level, got: {self.role}")

        return lib.wlr_xdg_toplevel_set_activated(self._ptr, activated)

    def set_tiled(self, tiled: int) -> int:
        return lib.wlr_xdg_toplevel_set_tiled(self._ptr, tiled)

    def set_fullscreen(self, fullscreened: bool) -> int:
        return lib.wlr_xdg_toplevel_set_fullscreen(self._ptr, fullscreened)

    def send_close(self) -> int:
        return lib.wlr_xdg_toplevel_send_close(self._ptr)

    def surface_at(
        self, surface_x: float, surface_y: float
    ) -> tuple[Surface | None, float, float]:
        """Find a surface within this xdg-surface tree at the given surface-local coordinates

        Returns the surface and coordinates in the leaf surface coordinate
        system or None if no surface is found at that location.
        """
        sub_x_data = ffi.new("double*")
        sub_y_data = ffi.new("double*")
        surface_ptr = lib.wlr_xdg_surface_surface_at(
            self._ptr, surface_x, surface_y, sub_x_data, sub_y_data
        )
        if surface_ptr == ffi.NULL:
            return None, 0.0, 0.0

        return Surface(surface_ptr), sub_x_data[0], sub_y_data[0]

    def for_each_surface(
        self, iterator: SurfaceCallback[T], data: T | None = None
    ) -> None:
        """Call iterator on each surface and popup in the xdg-surface tree

        Call `iterator` on each surface and popup in the xdg-surface tree, with
        the surface's position relative to the root xdg-surface. The function
        is called from root to leaves (in rendering order).

        :param iterator:
            The method that should be invoked
        :param data:
            The data that is passed as the last argument to the iterator method
        """
        py_handle = (iterator, data)
        handle = ffi.new_handle(py_handle)
        lib.wlr_xdg_surface_for_each_surface(
            self._ptr, lib.surface_iterator_callback, handle
        )


class XdgSurfaceConfigure(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_xdg_surface_configure *", ptr)

    @property
    def surface(self) -> XdgSurface:
        # TODO: keep weakref
        return XdgSurface(self._ptr.surface)

    @property
    def serial(self) -> int:
        return self._ptr.serial


class XdgTopLevel(Ptr):
    def __init__(self, ptr) -> None:
        """A top level surface object

        :param ptr:
            The wlr_xdg_toplevel cdata pointer
        """
        self._ptr = ptr

        self.request_maximize_event = Signal(
            ptr=ffi.addressof(self._ptr.events.request_maximize)
        )
        self.request_fullscreen_event = Signal(
            ptr=ffi.addressof(self._ptr.events.request_fullscreen),
            data_wrapper=XdgTopLevelSetFullscreenEvent,
        )
        self.request_minimize_event = Signal(
            ptr=ffi.addressof(self._ptr.events.request_minimize)
        )
        self.request_move_event = Signal(
            ptr=ffi.addressof(self._ptr.events.request_move),
            data_wrapper=XdgTopLevelMoveEvent,
        )
        self.request_resize_event = Signal(
            ptr=ffi.addressof(self._ptr.events.request_resize),
            data_wrapper=XdgTopLevelResizeEvent,
        )
        self.request_show_window_menu_event = Signal(
            ptr=ffi.addressof(self._ptr.events.request_show_window_menu),
            data_wrapper=XdgTopLevelShowWindowMenuEvent,
        )
        self.set_parent_event = Signal(ptr=ffi.addressof(self._ptr.events.set_parent))
        self.set_title_event = Signal(ptr=ffi.addressof(self._ptr.events.set_title))
        self.set_app_id_event = Signal(ptr=ffi.addressof(self._ptr.events.set_app_id))

    @property
    def parent(self) -> XdgSurface | None:
        """The surface of the parent of this toplevel"""
        parent_ptr = self._ptr.parent
        if parent_ptr is None:
            return None
        return XdgSurface(parent_ptr)

    @property
    def title(self) -> str | None:
        """The title of the toplevel object"""
        return str_or_none(self._ptr.title)

    @property
    def app_id(self) -> str | None:
        """The app id of the toplevel object"""
        return str_or_none(self._ptr.app_id)


class XdgTopLevelMoveEvent(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_xdg_toplevel_move_event *", ptr)

    @property
    def surface(self) -> Surface:
        # TODO: keep weakref
        return Surface(self._ptr.surface)

    # TODO: seat client

    @property
    def serial(self) -> int:
        return self._ptr.serial


class XdgTopLevelResizeEvent(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_xdg_toplevel_resize_event *", ptr)

    @property
    def surface(self) -> Surface:
        # TODO: keep weakref
        return Surface(self._ptr.surface)

    # TODO: seat client

    @property
    def serial(self) -> int:
        return self._ptr.serial

    @property
    def edges(self) -> Edges:
        return self._ptr.edges


class XdgTopLevelSetFullscreenEvent(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_xdg_toplevel_set_fullscreen_event *", ptr)

    @property
    def surface(self) -> Surface:
        # TODO: keep weakref
        return Surface(self._ptr.surface)

    @property
    def fullscreen(self) -> bool:
        return self._ptr.fullscreen

    @property
    def output(self) -> Output:
        return Output(self._ptr.output)


class XdgTopLevelShowWindowMenuEvent(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_xdg_toplevel_show_window_menu_event *", ptr)

    @property
    def surface(self) -> Surface:
        # TODO: keep weakref
        return Surface(self._ptr.surface)

    # TODO: seat client

    @property
    def serial(self) -> int:
        return self._ptr.serial

    @property
    def x(self) -> int:
        return self._ptr.x

    @property
    def y(self) -> int:
        return self._ptr.y


class XdgPopup(Ptr):
    def __init__(self, ptr) -> None:
        """A wlr_xdg_popup

        :param ptr:
            The wlr_xdg_popup cdata pointer
        """
        self._ptr = ffi.cast("struct wlr_xdg_popup *", ptr)

    @property
    def base(self) -> XdgSurface:
        """The xdg surface associated with the popup"""
        return XdgSurface(self._ptr.base)

    @property
    def parent(self) -> Surface:
        """Parent Surface."""
        parent = Surface(self.parent)
        _weakkeydict[parent] = self
        return parent

    def unconstrain_from_box(self, box: Box) -> None:
        """
        Set the geometry of this popup to unconstrain it according to its xdg-positioner
        rules. The box should be in the popup's root toplevel parent surface coordinate
        system.
        """
        lib.wlr_xdg_popup_unconstrain_from_box(self._ptr, box._ptr)



# Copyright (c) 2021 Matt Colligan
from __future__ import annotations

import enum
from dataclasses import dataclass
from typing import TYPE_CHECKING
from weakref import WeakKeyDictionary

from pywayland.server import Signal

from wlroots import ffi, PtrHasData, lib, Ptr
from .output import Output
from .surface import Surface
from .xdg_shell import SurfaceCallback, T

if TYPE_CHECKING:
    from pywayland.server import Display

_weakkeydict: WeakKeyDictionary = WeakKeyDictionary()


class LayerSurfaceV1KeyboardInteractivity(enum.IntEnum):
    NONE = 0
    EXCLUSIVE = 1
    ON_DEMAND = 2


class LayerSurfaceV1Anchor(enum.IntFlag):
    TOP = 1
    BOTTOM = 2
    LEFT = 4
    RIGHT = 8
    VERTICAL = TOP | BOTTOM
    HORIZONTAL = LEFT | RIGHT


class LayerShellV1Layer(enum.IntEnum):
    BACKGROUND = 0
    BOTTOM = 1
    TOP = 2
    OVERLAY = 3


@dataclass
class Margin:
    top: int
    right: int
    bottom: int
    left: int


class LayerSurfaceV1State(Ptr):
    def __init__(self, ptr):
        self._ptr = ptr

    @property
    def anchor(self) -> LayerSurfaceV1Anchor:
        return LayerSurfaceV1Anchor(self._ptr.anchor)

    @property
    def exclusive_zone(self) -> int:
        return self._ptr.exclusive_zone

    @property
    def margin(self) -> Margin:
        margin = self._ptr.margin
        return Margin(margin.top, margin.right, margin.bottom, margin.left)

    @property
    def keyboard_interactive(self) -> LayerSurfaceV1KeyboardInteractivity:
        return LayerSurfaceV1KeyboardInteractivity(self._ptr.keyboard_interactive)

    @property
    def desired_width(self) -> int:
        return self._ptr.desired_width

    @property
    def desired_height(self) -> int:
        return self._ptr.desired_height

    @property
    def actual_width(self) -> int:
        return self._ptr.actual_width

    @property
    def actual_height(self) -> int:
        return self._ptr.actual_height

    @property
    def layer(self) -> LayerShellV1Layer:
        return LayerShellV1Layer(self._ptr.layer)


class LayerSurfaceV1(PtrHasData):
    def __init__(self, ptr):
        self._ptr = ffi.cast("struct wlr_layer_surface_v1 *", ptr)

        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))
        self.map_event = Signal(ptr=ffi.addressof(self._ptr.events.map))
        self.unmap_event = Signal(ptr=ffi.addressof(self._ptr.events.unmap))
        self.new_popup_event = Signal(ptr=ffi.addressof(self._ptr.events.new_popup))

    @property
    def surface(self) -> Surface:
        surface_ptr = self._ptr.surface
        _weakkeydict[surface_ptr] = self._ptr
        return Surface(surface_ptr)

    @property
    def output(self) -> Output | None:
        output_ptr = self._ptr.output
        if output_ptr == ffi.NULL:
            return None
        _weakkeydict[output_ptr] = self._ptr
        return Output(output_ptr)

    @output.setter
    def output(self, output: Output):
        self._ptr.output = output._ptr

    @property
    def added(self) -> bool:
        return self._ptr.added

    @property
    def configured(self) -> bool:
        return self._ptr.configured

    @property
    def mapped(self) -> bool:
        return self._ptr.mapped

    @property
    def pending(self) -> LayerSurfaceV1State:
        state_ptr = self._ptr.pending
        _weakkeydict[state_ptr] = self._ptr
        return LayerSurfaceV1State(state_ptr)

    @property
    def current(self) -> LayerSurfaceV1State:
        state_ptr = self._ptr.current
        _weakkeydict[state_ptr] = self._ptr
        return LayerSurfaceV1State(state_ptr)

    def configure(self, width: int, height: int) -> None:
        """
        Notifies the layer surface to configure itself with this width/height. The
        layer_surface will signal its map event when the surface is ready to assume
        this size.
        """
        lib.wlr_layer_surface_v1_configure(self._ptr, width, height)

    def destroy(self) -> None:
        lib.wlr_layer_surface_v1_destroy(self._ptr)

    @classmethod
    def from_wlr_surface(cls, surface: Surface):
        surface_ptr = lib.wlr_layer_surface_v1_from_wlr_surface(surface._ptr)
        _weakkeydict[surface_ptr] = surface._ptr
        return LayerSurfaceV1(surface_ptr)

    def for_each_surface(
        self, iterator: SurfaceCallback[T], data: T | None = None
    ) -> None:
        """
        Calls the iterator function for each sub-surface and popup of this surface
        """
        py_handle = (iterator, data)
        handle = ffi.new_handle(py_handle)
        lib.wlr_layer_surface_v1_for_each_surface(
            self._ptr, lib.surface_iterator_callback, handle
        )

    def surface_at(self, sx: float, sy: float) -> tuple[Surface | None, float, float]:
        """
        Find a surface within this layer-surface tree at the given surface-local
        coordinates. Returns the surface and coordinates in the leaf surface
        coordinate system or None if no surface is found at that location.
        """
        sub_x_data = ffi.new("double*")
        sub_y_data = ffi.new("double*")
        surface_ptr = lib.wlr_layer_surface_v1_surface_at(
            self._ptr, sx, sy, sub_x_data, sub_y_data
        )
        if surface_ptr == ffi.NULL:
            return None, 0.0, 0.0

        return Surface(surface_ptr), sub_x_data[0], sub_y_data[0]


class LayerShellV1(PtrHasData):
    def __init__(self, display: Display) -> None:
        """Create an wlr_xdg_output_manager_v1"""
        self._ptr = lib.wlr_layer_shell_v1_create(display._ptr)

        self.new_surface_event = Signal(
            ptr=ffi.addressof(self._ptr.events.new_surface), data_wrapper=LayerSurfaceV1
        )
        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))



# Copyright (c) 2022

from __future__ import annotations

import enum

from pywayland.server import Display, Signal

from wlroots import Ptr, ffi, lib

from .output import Output


class OutputPowerManagementV1Mode(enum.IntEnum):
    OFF = 0
    ON = 1


class OutputPowerV1(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_output_power_v1 *", ptr)

    @property
    def output(self) -> Output:
        return Output(self._ptr.output)

    @property
    def manager(self) -> OutputPowerManagerV1:
        return OutputPowerManagerV1(self._ptr.manager)


class OutputPowerManagerV1(Ptr):
    def __init__(self, display: Display) -> None:
        self._ptr = lib.wlr_output_power_manager_v1_create(display._ptr)

        self.set_mode_event = Signal(
            ptr=ffi.addressof(self._ptr.events.set_mode),
            data_wrapper=OutputPowerV1SetModeEvent,
        )

        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))


class OutputPowerV1SetModeEvent(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_output_power_v1_set_mode_event *", ptr)

    @property
    def output(self) -> Output:
        return Output(self._ptr.output)

    @property
    def mode(self) -> OutputPowerManagementV1Mode:
        return OutputPowerManagementV1Mode(self._ptr.mode)



# Copyright (c) Antonin Riha 2022
from pywayland.server import Display, Signal

from wlroots import Ptr, ffi, lib, PtrHasData

from .surface import Surface


class IdleInhibitorManagerV1(Ptr):
    def __init__(self, display: Display) -> None:
        self._ptr = lib.wlr_idle_inhibit_v1_create(display._ptr)

        self.new_inhibitor_event = Signal(
            ptr=ffi.addressof(self._ptr.events.new_inhibitor),
            data_wrapper=IdleInhibitorV1,
        )
        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))


class IdleInhibitorV1(PtrHasData):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_idle_inhibitor_v1 *", ptr)

        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))

    @property
    def surface(self) -> Surface:
        return Surface(self._ptr.surface)



# Copyright (c) 2019 Sean Vig

from __future__ import annotations

import enum
import weakref

from pywayland.server import Display, Signal

from . import ffi, lib, Ptr
from wlroots.util.log import logger
from wlroots.wlr_types.input_device import InputDevice
from wlroots.wlr_types.output import Output


class BackendType(enum.Enum):
    AUTO = enum.auto()
    HEADLESS = enum.auto()


class Backend(Ptr):
    def __init__(self, display: Display, *, backend_type=BackendType.AUTO) -> None:
        """Create a backend to interact with a Wayland display

        :param display:
            The Wayland server display to create the backend against.  If this
            display is destroyed, the backend will be automatically cleaned-up.
        :param backend_type:
            The type of the backend to create.  The default (auto-detection)
            will use environment variables, including $DISPLAY (for X11 nested
            backends), $WAYLAND_DISPLAY (for Wayland backends), and
            $WLR_BACKENDS (to set the available backends).
        """
        if backend_type == BackendType.AUTO:
            ptr = lib.wlr_backend_autocreate(display._ptr)
        elif backend_type == BackendType.HEADLESS:
            ptr = lib.wlr_headless_backend_create(display._ptr)
        else:
            raise ValueError(f"Unknown backend type: {backend_type}")

        if ptr == ffi.NULL:
            raise RuntimeError("Failed to create wlroots backend")

        self._ptr = ffi.gc(ptr, lib.wlr_backend_destroy)
        self._weak_display = weakref.ref(display)

        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))
        self.new_input_event = Signal(
            ptr=ffi.addressof(self._ptr.events.new_input), data_wrapper=InputDevice
        )
        self.new_output_event = Signal(
            ptr=ffi.addressof(self._ptr.events.new_output), data_wrapper=Output
        )

    def destroy(self) -> None:
        """Destroy the backend and clean up all of its resources

        Normally called automatically when the wl_display is destroyed.
        """
        if self._ptr is None:
            logger.warning("Backend already destroyed, doing nothing")
        else:
            maybe_display = self._weak_display()
            if maybe_display is not None and maybe_display._ptr is not None:
                ffi.release(self._ptr)
            else:
                logger.warning(
                    "The display has already been cleaned up, clearing backend without destroying"
                )
                self._ptr = ffi.gc(self._ptr, None)

            self._ptr = None

    def start(self) -> None:
        """Start the backend

        This may signal new_input or new_output immediately, but may also wait
        until the display's event loop begins.
        """
        ret = lib.wlr_backend_start(self._ptr)
        if not ret:
            self.destroy()
            raise RuntimeError("Unable to start backend")

    def __enter__(self) -> Backend:
        """Context manager to create and clean-up the backend"""
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb) -> None:
        """Destroy the backend on context exit"""
        self.destroy()

    def get_session(self) -> Session:
        return Session(self)

    @property
    def is_headless(self) -> bool:
        return lib.wlr_backend_is_headless(self._ptr)


class Session:
    def __init__(self, backend: Backend) -> None:
        """The running session"""
        self._ptr = lib.wlr_backend_get_session(backend._ptr)

    def change_vt(self, vt: int) -> bool:
        return lib.wlr_session_change_vt(self._ptr, vt)



# Copyright (c) 2021 Matt Colligan

from weakref import WeakKeyDictionary

from pywayland.server import Display, Signal

from .input_device import InputDevice
from wlroots import ffi, lib, Ptr

_weakkeydict: WeakKeyDictionary = WeakKeyDictionary()


class VirtualKeyboardManagerV1(Ptr):
    def __init__(self, display: Display) -> None:
        """A wlr_virtual_keyboard_manager_v1 instance."""
        self._ptr = lib.wlr_virtual_keyboard_manager_v1_create(display._ptr)

        self.new_virtual_keyboard_event = Signal(
            ptr=ffi.addressof(self._ptr.events.new_virtual_keyboard),
            data_wrapper=VirtualKeyboardV1,
        )
        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))


class VirtualKeyboardV1(Ptr):
    def __init__(self, ptr) -> None:
        """A wlr_virtual_keyboard_v1 instance."""
        self._ptr = ffi.cast("struct wlr_virtual_keyboard_v1 *", ptr)

        self.destroy_event = Signal(
            ptr=ffi.addressof(self._ptr.events.destroy), data_wrapper=VirtualKeyboardV1
        )

    @property
    def input_device(self) -> InputDevice:
        device_ptr = ffi.addressof(self._ptr.input_device)
        _weakkeydict[device_ptr] = self._ptr
        return InputDevice(device_ptr)

    @property
    def has_keymap(self) -> bool:
        return self._ptr.has_keymap



# Copyright (c) Matt Colligan 2021

from __future__ import annotations

import enum
from typing import TYPE_CHECKING
from weakref import WeakKeyDictionary

from pywayland.server import Signal

from wlroots import ffi, PtrHasData, lib
from .surface import Surface

if TYPE_CHECKING:
    from pywayland.server import Display

_weakkeydict: WeakKeyDictionary = WeakKeyDictionary()


class XdgToplevelDecorationV1Mode(enum.IntEnum):
    NONE = lib.WLR_XDG_TOPLEVEL_DECORATION_V1_MODE_NONE
    CLIENT_SIDE = lib.WLR_XDG_TOPLEVEL_DECORATION_V1_MODE_CLIENT_SIDE
    SERVER_SIDE = lib.WLR_XDG_TOPLEVEL_DECORATION_V1_MODE_SERVER_SIDE


class XdgDecorationManagerV1(PtrHasData):
    def __init__(self, ptr) -> None:
        """An XDG decoration manager: wlr_xdg_decoration_manager_v1."""
        self._ptr = ffi.cast("struct wlr_xdg_decoration_manager_v1 *", ptr)

        self.new_toplevel_decoration_event = Signal(
            ptr=ffi.addressof(self._ptr.events.new_toplevel_decoration),
            data_wrapper=XdgToplevelDecorationV1,
        )
        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))

    @classmethod
    def create(cls, display: Display):
        """Create a wlr_xdg_decoration_manager_v1 for the given display."""
        ptr = lib.wlr_xdg_decoration_manager_v1_create(display._ptr)
        return cls(ptr)


class XdgToplevelDecorationV1(PtrHasData):
    def __init__(self, ptr) -> None:
        """struct wlr_xdg_toplevel_decoration_v1"""
        self._ptr = ffi.cast("struct wlr_xdg_toplevel_decoration_v1 *", ptr)

        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))
        self.request_mode_event = Signal(
            ptr=ffi.addressof(self._ptr.events.request_mode)
        )

    @property
    def surface(self) -> Surface:
        surface_ptr = self._ptr.surface
        _weakkeydict[surface_ptr] = self._ptr
        return Surface(surface_ptr)

    @property
    def manager(self) -> XdgDecorationManagerV1:
        manager_ptr = self._ptr.manager
        _weakkeydict[manager_ptr] = self._ptr
        return XdgDecorationManagerV1(manager_ptr)

    @property
    def added(self) -> bool:
        return self._ptr.added

    @property
    def current_mode(self) -> XdgToplevelDecorationV1Mode:
        return XdgToplevelDecorationV1Mode(self._ptr.current_mode)

    @property
    def client_pending_mode(self) -> XdgToplevelDecorationV1Mode:
        return XdgToplevelDecorationV1Mode(self._ptr.client_pending_mode)

    @property
    def server_pending_mode(self) -> XdgToplevelDecorationV1Mode:
        return XdgToplevelDecorationV1Mode(self._ptr.server_pending_mode)

    def set_mode(self, mode: XdgToplevelDecorationV1Mode) -> int:
        if mode == XdgToplevelDecorationV1Mode.NONE:
            raise ValueError("Toplevel decoration mode cannot be set to NONE.")
        return lib.wlr_xdg_toplevel_decoration_v1_set_mode(self._ptr, mode)



# Copyright (c) 2022 Sean Vig

from __future__ import annotations

from wlroots import lib, Ptr
from wlroots.renderer import Renderer
from wlroots.backend import Backend


class Allocator(Ptr):
    def __init__(self, ptr) -> None:
        """Create an allocator.

        The allocator is the bridge between the renderer and the backend. It
        handles the buffer creation, allowing wlroots to render onto the
        screen.
        """
        self._ptr = ptr

    @classmethod
    def autocreate(cls, backend: Backend, renderer: Renderer) -> Allocator:
        """Creates the adequate allocator given a backend and a renderer."""
        ret = lib.wlr_allocator_autocreate(backend._ptr, renderer._ptr)
        if not ret:
            raise RuntimeError("Unable to create an allocator.")
        return Allocator(ret)



from __future__ import annotations

import logging
import signal
from typing import TYPE_CHECKING, cast

from pywayland.protocol.wayland import WlKeyboard, WlSeat
from pywayland.server import Display, Listener
from wlroots import ffi, lib
from wlroots.allocator import Allocator
from wlroots.backend import Backend
from wlroots.renderer import Renderer
from wlroots.util.clock import Timespec
from wlroots.util.edges import Edges
from wlroots.util.log import logger
from wlroots.wlr_types import (
    Box,
    Cursor,
    Output,
    OutputLayout,
    Scene,
    SceneNode,
    Seat,
    Surface,
    XCursorManager,
    XdgShell,
)
from wlroots.wlr_types.cursor import WarpMode
from wlroots.wlr_types.input_device import ButtonState, InputDeviceType
from wlroots.wlr_types.keyboard import KeyboardModifier
from wlroots.wlr_types.pointer import (
    PointerEventButton,
    PointerEventMotion,
    PointerEventMotionAbsolute,
)
from wlroots.wlr_types.seat import RequestSetSelectionEvent
from wlroots.wlr_types.xdg_shell import XdgSurface, XdgSurfaceRole
from xkbcommon import xkb

from .cursor_mode import CursorMode
from .keyboard_handler import KeyboardHandler
from .view import View

if TYPE_CHECKING:
    from wlroots.wlr_types import InputDevice
    from wlroots.wlr_types.keyboard import KeyboardKeyEvent, KeyboardModifiers


def get_keysyms(xkb_state, keycode):
    syms_out = ffi.new("const xkb_keysym_t **")
    nsyms = lib.xkb_state_key_get_syms(xkb_state, keycode, syms_out)
    if nsyms > 0:
        assert syms_out[0] != ffi.NULL

    syms = [syms_out[0][i] for i in range(nsyms)]
    print(f"got {nsyms} syms: {syms}")
    return syms


class TinywlServer:
    def __init__(
        self,
        *,
        display: Display,
        backend: Backend,
        allocator: Allocator,
        renderer: Renderer,
        scene: Scene,
        xdg_shell: XdgShell,
        cursor: Cursor,
        cursor_manager: XCursorManager,
        seat: Seat,
        output_layout: OutputLayout,
    ) -> None:
        # elements that we need to hold on to
        self._display = display
        self._backend = backend
        self._allocator = allocator
        self._renderer = renderer
        self._scene = scene

        self._event_loop = self._display.get_event_loop()
        self._event_loop.add_signal(
            signal.SIGINT, self._terminate_signal_callback, self._display
        )

        # the xdg shell will generate new surfaces
        self._xdg_shell = xdg_shell
        self.views: list[View] = []

        # new pointing devices are attached to the cursor, and rendered with the manager
        self._cursor = cursor
        self._cursor_manager = cursor_manager

        # the seat manages the keyboard focus information
        self._seat = seat
        self.keyboards: list[KeyboardHandler] = []
        self.cursor_mode = CursorMode.PASSTHROUGH
        self.grabbed_view: View | None = None
        self.grab_x = 0.0
        self.grab_y = 0.0
        self.grab_geobox: Box | None = None
        self.resize_edges: Edges = Edges.NONE

        self._output_layout = output_layout
        self.outputs: list[Output] = []

        xdg_shell.new_surface_event.add(Listener(self.server_new_xdg_surface))

        backend.new_output_event.add(Listener(self.server_new_output))

        cursor.motion_event.add(Listener(self.server_cursor_motion))
        cursor.motion_absolute_event.add(Listener(self.server_cursor_motion_absolute))
        cursor.button_event.add(Listener(self.server_cursor_button))
        cursor.axis_event.add(Listener(self.server_cursor_axis))
        cursor.frame_event.add(Listener(self.server_cursor_frame))

        seat.request_set_cursor_event.add(Listener(self.seat_request_cursor))
        seat.request_set_selection_event.add(Listener(self.seat_request_set_selection))

        backend.new_input_event.add(Listener(self.server_new_input))

    def _terminate_signal_callback(self, sig_num: int, display: Display):
        logging.info("Terminating event loop.")
        display.terminate()

    def view_at(
        self, layout_x, layout_y
    ) -> tuple[View | None, Surface | None, float, float]:
        for view in self.views[::-1]:
            surface, x, y = view.view_at(layout_x, layout_y)
            if surface is not None:
                return view, surface, x, y
        return None, None, 0, 0

    def _process_cursor_move(self) -> None:
        # Move the grabbed view to the new position
        assert self.grabbed_view is not None
        x = self.grabbed_view.x = self._cursor.x - self.grab_x
        y = self.grabbed_view.y = self._cursor.y - self.grab_y
        self.grabbed_view.scene_node.set_position(int(x), int(y))

    def _process_cursor_resize(self) -> None:
        assert self.grabbed_view is not None
        assert self.grab_geobox is not None

        new_left = self.grab_geobox.x
        new_right = self.grab_geobox.x + self.grab_geobox.width
        new_top = self.grab_geobox.y
        new_bottom = self.grab_geobox.y + self.grab_geobox.height

        border_x = int(self._cursor.x - self.grab_x)
        border_y = int(self._cursor.y - self.grab_y)

        if self.resize_edges & Edges.TOP:
            new_top = min(border_y, new_bottom - 1)
        elif self.resize_edges & Edges.BOTTOM:
            new_bottom = max(border_y, new_top + 1)

        if self.resize_edges & Edges.LEFT:
            new_left = min(border_x, new_right - 1)
        elif self.resize_edges & Edges.RIGHT:
            new_right = max(border_x, new_left + 1)

        geo_box = self.grabbed_view.xdg_surface.get_geometry()
        self.grabbed_view.x = new_left - geo_box.x
        self.grabbed_view.y = new_top - geo_box.y

        new_width = new_right - new_left
        new_height = new_bottom - new_top

        self.grabbed_view.xdg_surface.set_size(new_width, new_height)

    def process_cursor_motion(self, time) -> None:
        if self.cursor_mode == CursorMode.MOVE:
            self._process_cursor_move()
            return
        elif self.cursor_mode == CursorMode.RESIZE:
            print("RESIZING")
            self._process_cursor_resize()
            return

        view, surface, sx, sy = self.view_at(self._cursor.x, self._cursor.y)
        logging.debug("Processing cursor motion: %s, %s", sx, sy)

        if view is None:
            self._cursor_manager.set_cursor_image("left_ptr", self._cursor)

        if surface is None:
            # Clear pointer focus so future button events and such are not sent
            # to the last client to have the cursor over it.
            self._seat.pointer_clear_focus()
        else:
            # Send pointer enter and motion events.
            self._seat.pointer_notify_enter(surface, sx, sy)
            self._seat.pointer_notify_motion(time, sx, sy)

    def send_modifiers(
        self, modifiers: KeyboardModifiers, input_device: InputDevice
    ) -> None:
        self._seat.set_keyboard(input_device)
        self._seat.keyboard_notify_modifiers(modifiers)

    def send_key(self, key_event: KeyboardKeyEvent, input_device: InputDevice) -> None:
        keyboard = input_device.keyboard
        keyboard_modifier = keyboard.modifier

        handled = False
        # If alt is held down and this button was _pressed_, we attempt to
        # process it as a compositor keybinding
        if (
            keyboard_modifier == KeyboardModifier.ALT
            and key_event.state == WlKeyboard.key_state.pressed
        ):
            # translate libinput keycode -> xkbcommon
            keycode = key_event.keycode + 8
            keysyms = get_keysyms(keyboard._ptr.xkb_state, keycode)

            for keysym in keysyms:
                if self.handle_keybinding(keysym):
                    handled = True
                    break

        # Otherwise, we pass it along to the client
        if not handled:
            self._seat.set_keyboard(input_device)
            self._seat.keyboard_notify_key(key_event)

    def handle_keybinding(self, keysym: int) -> bool:
        if keysym == xkb.keysym_from_name("Escape"):
            self._display.terminate()
        elif keysym == xkb.keysym_from_name("F1"):
            if len(self.views) >= 2:
                *rest, new_view, prev_view = self.views
                self.focus_view(new_view)
                self.views = [prev_view] + rest + [new_view]

        else:
            return False
        return True

    def focus_view(self, view: View, surface: Surface | None = None) -> None:
        """Focus a given XDG surface

        Moves the surface to the front of the list for rendering.  Sets the
        surface to receive keyboard events.
        """
        logging.info("focus surface")
        if surface is None:
            surface = view.xdg_surface.surface

        previous_surface = self._seat.keyboard_state.focused_surface
        if previous_surface == surface:
            # Don't re-focus an already focused surface
            logging.info("Focus unchanged, exiting")
            return

        if previous_surface is not None:
            # Deactivate the previously focused surface
            logging.info("Un-focusing previous")
            previous_xdg_surface = XdgSurface.from_surface(previous_surface)
            previous_xdg_surface.set_activated(False)

        view.scene_node.raise_to_top()
        # roll the given surface to the front of the list, copy and modify the
        # list, then save back to prevent any race conditions on list
        # modification
        views = self.views[:]
        views.remove(view)
        views.append(view)
        self.views = views
        # activate the new surface
        view.xdg_surface.set_activated(True)

        # Tell the seat to have the keyboard enter this surface. wlroots will
        # keep track of this and automatically send key events to the
        # appropriate clients without additional work on your part.
        keyboard = self._seat.keyboard
        self._seat.keyboard_notify_enter(view.xdg_surface.surface, keyboard)

    # #############################################################
    # surface handling callbacks

    def server_new_xdg_surface(self, listener, xdg_surface: XdgSurface) -> None:
        logger.info("new surface")

        if xdg_surface.role == XdgSurfaceRole.POPUP:
            # We must add xdg popups to the scene graph so they get rendered.
            # The wlroots scene graph provides a helper for this, but to use it
            # we must provide the proper parent scene node of the xdg popup. To
            # enable this, we always set the user data field of xdg_surfaces to
            # the corresponding scene node.
            parent_xdg_surface = XdgSurface.from_surface(xdg_surface.popup.parent)
            parent_scene_node = cast(SceneNode, parent_xdg_surface.data)
            scene_node = SceneNode.xdg_surface_create(parent_scene_node, xdg_surface)
            xdg_surface.data = scene_node
            return

        assert xdg_surface.role == XdgSurfaceRole.TOPLEVEL

        scene_node = SceneNode.xdg_surface_create(self._scene.node, xdg_surface)
        view = View(xdg_surface, self, scene_node)
        self.views.append(view)

    # #############################################################
    # output and frame handling callbacks

    def server_new_output(self, listener, output: Output) -> None:
        output.init_render(self._allocator, self._renderer)

        if output.modes != []:
            mode = output.preferred_mode()
            if mode is None:
                logger.error("Got no output mode")
                return
            output.set_mode(mode)
            output.enable()
            output.commit()

        self.outputs.append(output)
        self._output_layout.add_auto(output)

        output.frame_event.add(Listener(self.output_frame))

    def output_frame(self, listener, data) -> None:
        output = self.outputs[0]
        scene_output = self._scene.get_scene_output(output)
        scene_output.commit()

        now = Timespec.get_monotonic_time()
        scene_output.send_frame_done(now)

    # #############################################################
    # input handling callbacks

    def server_new_input(self, listener, input_device: InputDevice) -> None:
        if input_device.device_type == InputDeviceType.POINTER:
            self._server_new_pointer(input_device)
        elif input_device.device_type == InputDeviceType.KEYBOARD:
            self._server_new_keyboard(input_device)

        capabilities = WlSeat.capability.pointer
        if len(self.keyboards) > 0:
            capabilities |= WlSeat.capability.keyboard

        logging.info("new input %s", input_device.device_type)
        logging.info("capabilities %s", capabilities)

        self._seat.set_capabilities(capabilities)

    def _server_new_pointer(self, input_device: InputDevice) -> None:
        self._cursor.attach_input_device(input_device)

    def _server_new_keyboard(self, input_device: InputDevice) -> None:
        keyboard = input_device.keyboard

        xkb_context = xkb.Context()
        keymap = xkb_context.keymap_new_from_names()

        keyboard.set_keymap(keymap)
        keyboard.set_repeat_info(25, 600)

        keyboard_handler = KeyboardHandler(keyboard, input_device, self)
        self.keyboards.append(keyboard_handler)

        self._seat.set_keyboard(input_device)

    # #############################################################
    # cursor motion callbacks

    def server_cursor_motion(self, listener, event_motion: PointerEventMotion) -> None:
        logging.debug("cursor motion")
        # self._cursor.move(event_motion_absolute.device
        self._cursor.move(
            event_motion.delta_x, event_motion.delta_y, input_device=event_motion.device
        )
        self.process_cursor_motion(event_motion.time_msec)

    def server_cursor_motion_absolute(
        self, listener, event_motion_absolute: PointerEventMotionAbsolute
    ) -> None:
        logging.debug("cursor abs motion")
        self._cursor.warp(
            WarpMode.AbsoluteClosest,
            event_motion_absolute.x,
            event_motion_absolute.y,
            input_device=event_motion_absolute.device,
        )
        self.process_cursor_motion(event_motion_absolute.time_msec)

    def server_cursor_button(self, listener, event: PointerEventButton) -> None:
        logging.info("Got button click event %s", event.button_state)
        self._seat.pointer_notify_button(
            event.time_msec, event.button, event.button_state
        )

        view, surface, _, _ = self.view_at(self._cursor.x, self._cursor.y)
        if event.button_state == ButtonState.RELEASED:
            # exit interactive move/resize
            self.cursor_mode = CursorMode.PASSTHROUGH
        elif view is not None:
            self.focus_view(view, surface)

    def server_cursor_axis(self, listener, event) -> None:
        self._seat.pointer_notify_axis(
            event.time_msec,
            event.orientation,
            event.delta,
            event.delta_discrete,
            event.source,
        )

    def server_cursor_frame(self, listener, data) -> None:
        self._seat.pointer_notify_frame()

    # #############################################################
    # seat callbacks

    def seat_request_cursor(self, listener, event):
        # This event is rasied by the seat when a client provides a cursor image
        # TODO: check that seat client is correct
        self._cursor.set_surface(event.surface, event.hotspot)

    def seat_request_set_selection(
        self, listener, event: RequestSetSelectionEvent
    ) -> None:
        print("request set selection")
        self._seat.set_selection(event._ptr.source, event.serial)



# Copyright (c) Sean Vig 2019
# Copyright (c) Matt Colligan 2021

from __future__ import annotations

from typing import TYPE_CHECKING

from pywayland.server import Signal
from pywayland.protocol.wayland import WlOutput

from wlroots import ffi, PtrHasData, lib, Ptr, str_or_none
from wlroots.util.region import PixmanRegion32
from .matrix import Matrix

if TYPE_CHECKING:
    from wlroots.allocator import Allocator
    from wlroots.renderer import Renderer


class Output(PtrHasData):
    def __init__(self, ptr) -> None:
        """A compositor output region

        This typically corresponds to a monitor that displays part of the
        compositor space.

        The `frame` event will be emitted when it is a good time for the
        compositor to submit a new frame.

        To render a new frame, compositors should call
        `wlr_output_attach_render`, render and call `wlr_output_commit`. No
        rendering should happen outside a `frame` event handler or before
        `wlr_output_attach_render`.

        :param ptr:
            The wlr_output cdata pointer
        """
        self._ptr = ffi.cast("struct wlr_output *", ptr)

        self.frame_event = Signal(ptr=ffi.addressof(self._ptr.events.frame))
        self.damage_event = Signal(ptr=ffi.addressof(self._ptr.events.damage))
        self.needs_frame_event = Signal(ptr=ffi.addressof(self._ptr.events.needs_frame))
        self.precommit_event = Signal(ptr=ffi.addressof(self._ptr.events.precommit))
        self.commit_event = Signal(ptr=ffi.addressof(self._ptr.events.commit))
        self.present_event = Signal(ptr=ffi.addressof(self._ptr.events.present))
        self.bind_event = Signal(ptr=ffi.addressof(self._ptr.events.bind))
        self.enable_event = Signal(ptr=ffi.addressof(self._ptr.events.enable))
        self.mode_event = Signal(ptr=ffi.addressof(self._ptr.events.mode))
        self.description_event = Signal(ptr=ffi.addressof(self._ptr.events.description))
        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))

    @property
    def name(self) -> str | None:
        """The name of the output"""
        return str_or_none(self._ptr.name)

    @property
    def description(self) -> str | None:
        """The description of the output"""
        return str_or_none(self._ptr.description)

    @property
    def make(self) -> str | None:
        return str_or_none(self._ptr.make)

    @property
    def model(self) -> str | None:
        return str_or_none(self._ptr.model)

    @property
    def serial(self) -> str | None:
        return str_or_none(self._ptr.serial)

    @property
    def physical_size_mm(self) -> tuple[int, int]:
        """Returns the width and height of the output, in millimeters"""
        return self._ptr.phys_width, self._ptr.phys_height

    @property
    def modes(self):
        if lib.wl_list_empty(ffi.addressof(self._ptr.modes)) == 1:
            return []

    @property
    def enabled(self) -> bool:
        return self._ptr.enabled

    @property
    def current_mode(self) -> OutputMode:
        return OutputMode(self._ptr.current_mode)

    @property
    def scale(self) -> float:
        return self._ptr.scale

    @property
    def transform(self) -> WlOutput.transform:
        return WlOutput.transform(self._ptr.transform)

    @property
    def transform_matrix(self) -> Matrix:
        """The transform matrix giving the projection of the output"""
        return Matrix(self._ptr.transform_matrix)

    def enable(self, *, enable: bool = True) -> None:
        """Enables or disables the output

        A disabled output is turned off and doesn't emit `frame` events.
        """
        lib.wlr_output_enable(self._ptr, enable)

    def preferred_mode(self) -> OutputMode | None:
        """Returns the preferred mode for this output

        If the output doesn't support modes, returns None.
        """
        output_mode_ptr = lib.wlr_output_preferred_mode(self._ptr)
        if output_mode_ptr == ffi.NULL:
            return None

        return OutputMode(output_mode_ptr)

    def set_mode(self, mode: OutputMode) -> None:
        """Sets the output mode

        The output needs to be enabled.
        """
        lib.wlr_output_set_mode(self._ptr, mode._ptr)

    def set_custom_mode(self, width: int, height: int, refresh: int) -> None:
        """
        Sets a custom mode on the output. If modes are available, they are preferred.
        Setting `refresh` to zero lets the backend pick a preferred value. The
        output needs to be enabled.
        """
        lib.wlr_output_set_custom_mode(self._ptr, width, height, refresh)

    def create_global(self) -> None:
        """Create the global corresponding to the output"""
        lib.wlr_output_create_global(self._ptr)

    def __enter__(self) -> Output:
        """Start rendering frame"""
        return self

    def __exit__(self, exc_type, exc_value, exc_tb) -> None:
        """Stop rendering frame, commit when exiting normally, otherwise rollback"""
        if exc_type is None:
            self.commit()
        else:
            self.rollback()

    def init_render(self, allocator: Allocator, renderer: Renderer) -> None:
        """Initialize the output's rendering subsystem with the provided allocator and renderer.

        Can only be called once.

        Call this function prior to any call to `attach_render`, `commit`, or
        `cursor_create`. The buffer capabilities of the provided must match the
        capabilities of the output's backend.  Raises error otherwise.
        """
        if not lib.wlr_output_init_render(self._ptr, allocator._ptr, renderer._ptr):
            raise RuntimeError(
                "Output capabilities must match the capabilities of the output's backend."
            )

    def attach_render(self) -> None:
        """Attach the renderer's buffer to the output

        Compositors must call this function before rendering. After they are
        done rendering, they should call `.commit()` to submit the new frame.
        """
        if not lib.wlr_output_attach_render(self._ptr, ffi.NULL):
            raise RuntimeError("Unable to attach render")

    def commit(self) -> None:
        """Commit the pending output state

        If `.attach_render` has been called, the pending frame will be
        submitted for display.
        """
        if not lib.wlr_output_test(self._ptr):
            self.rollback()
            raise RuntimeError("Rendering on output failed")
        if not lib.wlr_output_commit(self._ptr):
            raise RuntimeError("Unable to commit output")

    def rollback(self) -> None:
        """Discard the pending output state"""
        lib.wlr_output_rollback(self._ptr)

    def effective_resolution(self) -> tuple[int, int]:
        """Computes the transformed and scaled output resolution"""
        width_ptr = ffi.new("int *")
        height_ptr = ffi.new("int *")
        lib.wlr_output_effective_resolution(self._ptr, width_ptr, height_ptr)
        width = width_ptr[0]
        height = height_ptr[0]
        return width, height

    def transformed_resolution(self) -> tuple[int, int]:
        """Computes the transformed output resolution"""
        width_ptr = ffi.new("int *")
        height_ptr = ffi.new("int *")
        lib.wlr_output_transformed_resolution(self._ptr, width_ptr, height_ptr)
        width = width_ptr[0]
        height = height_ptr[0]
        return width, height

    def render_software_cursors(self, damage: PixmanRegion32 | None = None) -> None:
        """Renders software cursors

        This is a utility function that can be called when compositors render.
        """
        if damage is None:
            lib.wlr_output_render_software_cursors(self._ptr, ffi.NULL)
        else:
            lib.wlr_output_render_software_cursors(self._ptr, damage._ptr)

    @staticmethod
    def transform_invert(transform: WlOutput.transform) -> WlOutput.transform:
        """Returns the transform that, when composed with transform gives transform.normal"""
        return WlOutput.transform(lib.wlr_output_transform_invert(transform))

    def set_damage(self, damage: PixmanRegion32) -> None:
        """
        Set the damage region for the frame to be submitted. This is the region of
        the screen that has changed since the last frame.

        Compositors implementing damage tracking should call this function with the
        damaged region in output-buffer-local coordinates.

        This region is not to be confused with the renderer's buffer damage, ie. the
        region compositors need to repaint. Compositors usually need to repaint more
        than what changed since last frame since multiple render buffers are used.
        """
        lib.wlr_output_set_damage(self._ptr, damage._ptr)

    def set_transform(self, transform: WlOutput.transform) -> None:
        """
        Sets a transform for the output.

        Transform is double-buffered state, see `wlr_output_commit`.
        """
        lib.wlr_output_set_transform(self._ptr, transform)

    def set_scale(self, scale: float) -> None:
        """
        Sets a scale for the output.

        Scale is double-buffered state, see `wlr_output_commit`.
        """
        lib.wlr_output_set_scale(self._ptr, scale)

    def test(self) -> bool:
        """
        Test whether the pending output state would be accepted by the backend. If
        this function returns true, `wlr_output_commit` can only fail due to a
        runtime error.

        This function doesn't mutate the pending state.
        """
        return lib.wlr_output_test(self._ptr)

    @property
    def is_headless(self) -> bool:
        return lib.wlr_output_is_headless(self._ptr)


class OutputMode(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ptr

    @property
    def width(self) -> int:
        return self._ptr.width

    @property
    def height(self) -> int:
        return self._ptr.height

    @property
    def refresh_mhz(self) -> int:
        return self._ptr.refresh

    @property
    def preferred(self) -> int:
        return self._ptr.preferred



# Copyright (c) 2022 Aakash Sen Sharma

from pywayland.server import Display, Signal
from wlroots import Ptr, ffi, lib


class ExportDmabufManagerV1(Ptr):
    def __init__(self, display: Display) -> None:
        """A `struct wlr_export_dmabuf_manager_v1`

        :param display:
            The wayland display
        """

        self._ptr = lib.wlr_export_dmabuf_manager_v1_create(display._ptr)
        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))



# Copyright (c) 2022 Matt Colligan

from __future__ import annotations

from typing import Iterable
from weakref import WeakKeyDictionary

from pywayland.protocol.wayland.wl_pointer import WlPointer
from pywayland.server import Display, Signal

from wlroots import Ptr, ffi, lib
from wlroots.wlr_types.input_device import InputDevice
from wlroots.wlr_types.pointer import PointerEventAxis

_weakkeydict: WeakKeyDictionary = WeakKeyDictionary()


class VirtualPointerManagerV1(Ptr):
    def __init__(self, display: Display) -> None:
        """A wlr_virtual_pointer_manager_v1 struct."""
        self._ptr = lib.wlr_virtual_pointer_manager_v1_create(display._ptr)

        self.new_virtual_pointer_event = Signal(
            ptr=ffi.addressof(self._ptr.events.new_virtual_pointer),
            data_wrapper=VirtualPointerV1NewPointerEvent,
        )
        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))


class VirtualPointerV1NewPointerEvent(Ptr):
    def __init__(self, ptr) -> None:
        """A wlr_virtual_pointer_v1_new_pointer_event struct."""
        self._ptr = ffi.cast("struct wlr_virtual_pointer_v1_new_pointer_event *", ptr)

    @property
    def new_pointer(self) -> VirtualPointerV1:
        return VirtualPointerV1(self._ptr.new_pointer)


class VirtualPointerV1(Ptr):
    def __init__(self, ptr) -> None:
        """A wlr_virtual_pointer_v1 struct."""
        self._ptr = ffi.cast("struct wlr_virtual_pointer_v1 *", ptr)

        self.destroy_event = Signal(
            ptr=ffi.addressof(self._ptr.events.destroy), data_wrapper=VirtualPointerV1
        )

    @property
    def input_device(self) -> InputDevice:
        device_ptr = ffi.addressof(self._ptr.input_device)
        _weakkeydict[device_ptr] = self._ptr
        return InputDevice(device_ptr)

    @property
    def axis_event(self) -> Iterable[PointerEventAxis]:
        for axis_event in self._ptr.axis_event:
            yield PointerEventAxis(axis_event)

    @property
    def axis(self) -> WlPointer.axis:
        return WlPointer.axis(self._ptr.axis)

    @property
    def axis_valid(self) -> Iterable[bool]:
        yield from self._ptr.axis_valid



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from pywayland.server import Listener
from wlroots.util.box import Box
from wlroots.util.edges import Edges

from .cursor_mode import CursorMode

if TYPE_CHECKING:
    from wlroots.wlr_types import SceneNode, Surface
    from wlroots.wlr_types.xdg_shell import XdgSurface

    from .server import TinywlServer


class View:
    def __init__(
        self,
        xdg_surface: XdgSurface,
        tinywl_server: TinywlServer,
        scene_node: SceneNode,
    ) -> None:
        self.xdg_surface = xdg_surface
        self.tinywl_server = tinywl_server
        self.scene_node = scene_node
        self.mapped = False
        self.x = 0.0
        self.y = 0.0

        xdg_surface.map_event.add(Listener(self.xdg_toplevel_map))
        xdg_surface.unmap_event.add(Listener(self.xdg_toplevel_unmap))
        xdg_surface.destroy_event.add(Listener(self.xdg_toplevel_destroy))

        toplevel = xdg_surface.toplevel
        toplevel.request_move_event.add(Listener(self.xdg_toplevel_request_move))
        toplevel.request_resize_event.add(Listener(self.xdg_toplevel_request_resize))

    def xdg_toplevel_map(self, listener, data) -> None:
        logging.info("mapped new view")
        self.mapped = True
        self.tinywl_server.focus_view(self)

    def xdg_toplevel_unmap(self, listener, data) -> None:
        logging.info("unmapped view")
        self.mapped = False

    def xdg_toplevel_destroy(self, listener, data) -> None:
        logging.info("destroyed view")
        self.tinywl_server.views.remove(self)

    def xdg_toplevel_request_move(self, listener, data) -> None:
        # This event is raised when a client would like to begin an interactive
        # move, typically because the user clicked on their client-side
        # decorations. Note that a more sophisticated compositor should check
        # the provied serial against a list of button press serials sent to
        # this client, to prevent the client from requesting this whenever they
        # want.
        logging.info("request move start")
        self._begin_interactive(CursorMode.MOVE, Edges.NONE)

    def xdg_toplevel_request_resize(self, listener, event) -> None:
        # This event is raised when a client would like to begin an interactive
        # resize, typically because the user clicked on their client-side
        # decorations. Note that a more sophisticated compositor should check
        # the provied serial against a list of button press serials sent to
        # this client, to prevent the client from requesting this whenever they
        # want.
        logging.info("request resize start")
        self._begin_interactive(CursorMode.RESIZE, event.edges)

    def _begin_interactive(self, cursor_mode: CursorMode, edges: Edges) -> None:
        """This function sets up an interactive move or resize operation"""
        focused_surface = self.tinywl_server._seat.pointer_state.focused_surface
        if self.xdg_surface.surface != focused_surface:
            # Deny move/resize requests from unfocused clients
            logging.info("Denied begin interactive")
            # TODO: this doesn't seem to be correct for alacritty, should return here

        self.tinywl_server.grabbed_view = self
        self.tinywl_server.cursor_mode = cursor_mode

        if cursor_mode == CursorMode.MOVE:
            self.tinywl_server.grab_x = self.tinywl_server._cursor.x - self.x
            self.tinywl_server.grab_y = self.tinywl_server._cursor.y - self.y
        elif cursor_mode == CursorMode.RESIZE:
            box = self.xdg_surface.get_geometry()

            border_x = self.x + box.x + (box.width if edges & Edges.RIGHT else 0)
            border_y = self.y + box.y + (box.height if edges & Edges.BOTTOM else 0)

            self.tinywl_server.grab_x = self.tinywl_server._cursor.x - border_x
            self.tinywl_server.grab_y = self.tinywl_server._cursor.y - border_y
            self.tinywl_server.grab_geobox = Box(
                box.x + int(self.x), box.y + int(self.y), box.width, box.height
            )
            self.tinywl_server.resize_edges = edges

    def view_at(
        self, layout_x: int, layout_y: int
    ) -> tuple[Surface | None, float, float]:
        view_x = layout_x - self.x
        view_y = layout_y - self.y
        return self.xdg_surface.surface_at(view_x, view_y)



# Copyright Sean Vig (c) 2020

from __future__ import annotations

from weakref import WeakKeyDictionary

from pywayland.protocol.wayland import WlOutput
from pywayland.server import Signal

from wlroots import ffi, PtrHasData, lib, Ptr
from wlroots.util.clock import Timespec
from .texture import Texture

_weakkeydict: WeakKeyDictionary = WeakKeyDictionary()


class Surface(PtrHasData):
    def __init__(self, ptr):
        """Create a wlroots Surface

        :param ptr:
            The cdata for the given surface
        """
        self._ptr = ptr

        self.commit_event = Signal(ptr=ffi.addressof(self._ptr.events.commit))
        self.new_subsurface_event = Signal(
            ptr=ffi.addressof(self._ptr.events.new_subsurface),
            data_wrapper=SubSurface,
        )
        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))

    def __eq__(self, other: object) -> bool:
        """Check if the other surface is equal by the cdata pointer"""
        if not isinstance(other, Surface):
            return NotImplemented

        return self._ptr == other._ptr

    @property
    def is_xdg_surface(self) -> bool:
        """True if the current surface is an XDG surface"""
        return lib.wlr_surface_is_xdg_surface(self._ptr)

    @property
    def is_layer_surface(self) -> bool:
        """True if the current surface is a layer surface"""
        return lib.wlr_surface_is_layer_surface(self._ptr)

    @property
    def is_xwayland_surface(self) -> bool:
        """
        True if the current surface is an XWayland surface.

        Requires that pywlroots was built with XWayland support.
        """
        return lib.wlr_surface_is_xwayland_surface(self._ptr)

    @property
    def sx(self) -> int:
        """Surface local buffer x position"""
        return self._ptr.sx

    @property
    def sy(self) -> int:
        """Surface local buffer y position"""
        return self._ptr.sy

    @property
    def current(self) -> SurfaceState:
        """The current commited surface state"""
        current_ptr = self._ptr.current
        _weakkeydict[current_ptr] = self._ptr
        return SurfaceState(current_ptr)

    @property
    def previous(self) -> SurfaceState:
        """The state of the previous commit"""
        previous_ptr = self._ptr.previous
        _weakkeydict[previous_ptr] = self._ptr
        return SurfaceState(previous_ptr)

    def get_texture(self):
        """Get the texture of the buffer currently attached to this surface

        Returns None if no buffer is currently attached or if something went
        wrong with uploading the buffer.
        """
        texture_ptr = lib.wlr_surface_get_texture(self._ptr)
        if texture_ptr == ffi.NULL:
            return None

        return Texture(texture_ptr)

    def send_frame_done(self, when: Timespec) -> None:
        """Send a frame done event to the surface"""
        lib.wlr_surface_send_frame_done(self._ptr, when._ptr)


class SurfaceState(Ptr):
    def __init__(self, ptr):
        """The state of a given surface

        :param ptr:
            The cdata of the given surface state.
        """
        self._ptr = ptr

    @property
    def transform(self) -> WlOutput.transform:
        """Get the transform for the selected surface"""
        return WlOutput.transform(self._ptr.transform)

    @property
    def width(self) -> int:
        """In surface local width"""
        return self._ptr.width

    @property
    def height(self) -> int:
        """In surface local height"""
        return self._ptr.height


class SubSurface(PtrHasData):
    def __init__(self, ptr):
        """A wlroots subsurface

        :param ptr:
            The cdata for the given subsurface
        """
        self._ptr = ffi.cast("struct wlr_subsurface *", ptr)

        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))
        self.map_event = Signal(ptr=ffi.addressof(self._ptr.events.map))
        self.unmap_event = Signal(ptr=ffi.addressof(self._ptr.events.unmap))

    @property
    def surface(self) -> Surface:
        """Get the wlr_surface underlying this subsurface"""
        return Surface(self._ptr.surface)



# Copyright (c) Sean Vig 2018

import sys

from distutils.command.build import build
from setuptools import setup
from setuptools.command.install import install


NO_SETUP_REQUIRES_ARGUMENTS = ["--help", "-h", "sdist"]


class DummyBuild(build):
    def run(self):
        raise RuntimeError("Requested running build")


class DummyInstall(install):
    def run(self):
        raise RuntimeError("Requested running install")


def keywords_with_side_effects(argv):
    if any(arg in NO_SETUP_REQUIRES_ARGUMENTS for arg in argv):
        return {"cmdclass": {"build": DummyBuild, "install": DummyInstall}}
    else:
        return {
            "setup_requires": ["cffi>=1.12.0", "pywayland>=0.1.1", "xkbcommon>=0.2"],
            "cffi_modules": ["wlroots/ffi_build.py:ffi_builder"],
        }


sys.path.insert(0, "wlroots")

setup(**keywords_with_side_effects(sys.argv))



# Copyright (c) 2019 Sean Vig

from __future__ import annotations

from weakref import WeakKeyDictionary

from pywayland.server import Display, Signal
from pywayland.protocol.wayland import WlSeat

from wlroots import ffi, PtrHasData, lib, Ptr
from .data_device_manager import Drag
from .input_device import ButtonState, InputDevice
from .keyboard import Keyboard, KeyboardModifiers, KeyboardKeyEvent
from .pointer import AxisSource, AxisOrientation
from .surface import Surface

_weakkeydict: WeakKeyDictionary = WeakKeyDictionary()


class KeyboardGrab(Ptr):
    def __init__(self, seat: Seat) -> None:
        """Setup the keyboard grab"""
        self._ptr = ffi.new("struct wlr_seat_keyboard_grab *")
        self._seat = seat

    def __enter__(self) -> KeyboardGrab:
        """State the keyboard grab"""
        lib.wlr_seat_keyboard_start_grab(self._seat._ptr, self._ptr)
        return self

    def __exit__(self, exc_type, exc_value, exc_tb) -> None:
        """End the grab of the keyboard of this seat"""
        lib.wlr_seat_keyboard_end_grab(self._seat)


class Seat(PtrHasData):
    def __init__(self, display: Display, name: str) -> None:
        """Allocates a new seat and adds a seat global to the display

        :param display:
            The Wayland server display to attach the seat to
        :param name:
            The name of the seat to create
        """
        ptr = lib.wlr_seat_create(display._ptr, name.encode())
        self._ptr = ffi.gc(ptr, lib.wlr_seat_destroy)

        self.pointer_grab_begin_event = Signal(
            ptr=ffi.addressof(self._ptr.events.pointer_grab_begin)
        )
        self.pointer_grab_end_event = Signal(
            ptr=ffi.addressof(self._ptr.events.pointer_grab_end)
        )

        self.keyboard_grab_begin_event = Signal(
            ptr=ffi.addressof(self._ptr.events.keyboard_grab_begin)
        )
        self.keyboard_grab_end_event = Signal(
            ptr=ffi.addressof(self._ptr.events.keyboard_grab_end)
        )

        self.touch_grab_begin_event = Signal(
            ptr=ffi.addressof(self._ptr.events.touch_grab_begin)
        )
        self.touch_grab_end_event = Signal(
            ptr=ffi.addressof(self._ptr.events.touch_grab_end)
        )

        self.request_set_cursor_event = Signal(
            ptr=ffi.addressof(self._ptr.events.request_set_cursor),
            data_wrapper=PointerRequestSetCursorEvent,
        )

        # Called when an application _wants_ to set the selection
        self.request_set_selection_event = Signal(
            ptr=ffi.addressof(self._ptr.events.request_set_selection),
            data_wrapper=RequestSetSelectionEvent,
        )
        # Called after the data source is set for the selection
        self.set_selection_event = Signal(
            ptr=ffi.addressof(self._ptr.events.set_selection)
        )

        # Called when an application _wants_ to set the primary selection (user selects some data)
        self.request_set_primary_selection_event = Signal(
            ptr=ffi.addressof(self._ptr.events.request_set_primary_selection),
            data_wrapper=RequestSetPrimarySelectionEvent,
        )
        # Called after the primary selection source object is set
        self.set_primary_selection_event = Signal(
            ptr=ffi.addressof(self._ptr.events.set_primary_selection)
        )

        self.request_start_drag_event = Signal(
            ptr=ffi.addressof(self._ptr.events.request_start_drag),
            data_wrapper=RequestStartDragEvent,
        )
        self.start_drag_event = Signal(
            ptr=ffi.addressof(self._ptr.events.start_drag),
            data_wrapper=Drag,
        )

    @property
    def pointer_state(self) -> SeatPointerState:
        """The pointer state associated with the seat"""
        pointer_state_ptr = ffi.addressof(self._ptr.pointer_state)
        _weakkeydict[pointer_state_ptr] = self._ptr
        return SeatPointerState(pointer_state_ptr)

    @property
    def keyboard_state(self) -> SeatKeyboardState:
        """The keyboard state associated with the seat"""
        keyboard_state_ptr = ffi.addressof(self._ptr.keyboard_state)
        _weakkeydict[keyboard_state_ptr] = self._ptr
        return SeatKeyboardState(keyboard_state_ptr)

    @property
    def keyboard(self) -> Keyboard:
        """Get the keyboard associated with this seat"""
        keyboard_ptr = lib.wlr_seat_get_keyboard(self._ptr)
        return Keyboard(keyboard_ptr)

    def destroy(self) -> None:
        """Clean up the seat"""
        if self._ptr is not None:
            ffi.release(self._ptr)
            self._ptr = None

    @property
    def destroyed(self) -> bool:
        """Whether this seat has been destroyed."""
        return self._ptr is None

    def set_capabilities(self, capabilities: WlSeat.capability) -> None:
        """Updates the capabilities available on this seat

        Will automatically send them to all clients.

        :param capabilities:
            The Wayland seat capabilities to set on the seat.
        """
        lib.wlr_seat_set_capabilities(self._ptr, capabilities)

    def set_name(self, name: str) -> None:
        """Updates the name of this seat

        Will automatically send it to all clients.
        """
        lib.wlr_seat_set_name(self._ptr, name.encode())

    def pointer_surface_has_focus(self, surface: Surface) -> bool:
        """Whether or not the surface has pointer focus"""
        return lib.wlr_seat_pointer_surface_has_focus(self._ptr, surface._ptr)

    def pointer_clear_focus(self) -> None:
        """Clear the focused surface for the pointer and leave all entered surfaces.

        This function does not respect pointer grabs: you probably want
        `pointer_notify_clear_focus()` instead.
        """

        return lib.wlr_seat_pointer_clear_focus(self._ptr)

    def pointer_notify_clear_focus(self) -> None:
        """Notify the seat of a pointer leave event to the currently-focused surface.

        Defers to any grab of the pointer.
        """
        return lib.wlr_seat_pointer_notify_clear_focus(self._ptr)

    def pointer_notify_enter(
        self, surface: Surface, surface_x: float, surface_y: float
    ) -> None:
        """Notify the seat of a pointer enter event to the given surface

        Notify the seat of a pointer enter event to the given surface and
        request it to be the focused surface for the pointer. Pass
        surface-local coordinates where the enter occurred.
        """
        lib.wlr_seat_pointer_notify_enter(self._ptr, surface._ptr, surface_x, surface_y)

    def pointer_notify_motion(
        self, time_msec: int, surface_x: float, surface_y: float
    ) -> None:
        """Notify the seat of motion over the given surface

        Pass surface-local coordinates where the pointer motion occurred.
        """
        lib.wlr_seat_pointer_notify_motion(self._ptr, time_msec, surface_x, surface_y)

    def pointer_notify_button(
        self, time_msec: int, button: int, button_state: ButtonState
    ) -> int:
        """Notify the seat that a button has been pressed

        Returns the serial of the button press or zero if no button press was
        sent.
        """
        return lib.wlr_seat_pointer_notify_button(
            self._ptr, time_msec, button, button_state.value
        )

    def pointer_notify_axis(
        self,
        time_msec: int,
        orientation: AxisOrientation,
        value: float,
        value_discrete: int,
        source: AxisSource,
    ) -> None:
        """Notify the seat of an axis event"""
        lib.wlr_seat_pointer_notify_axis(
            self._ptr, time_msec, orientation.value, value, value_discrete, source.value
        )

    def pointer_notify_frame(self) -> None:
        """Notify the seat of a frame event

        Frame events are sent to end a group of events that logically belong
        together. Motion, button and axis events should all be followed by a
        frame event.
        """
        lib.wlr_seat_pointer_notify_frame(self._ptr)

    def pointer_has_grab(self):
        """Whether or not the pointer has a grab other than the default grab"""
        lib.wlr_seat_pointer_has_grab(self._ptr)

    def set_keyboard(self, input_device: InputDevice) -> None:
        """Set this keyboard as the active keyboard for the seat

        :param input_device:
            The input device associated to the keyboard to set
        """
        lib.wlr_seat_set_keyboard(self._ptr, input_device._ptr)

    def grab(self) -> KeyboardGrab:
        """Start a grab of the keyboard of this seat"""
        return KeyboardGrab(self)

    @property
    def has_grab(self) -> bool:
        """Whether or not the keyboard has a grab other than the default grab"""
        return lib.wlr_seat_keyboard_has_grab(self._ptr)

    def keyboard_notify_key(self, key_event: KeyboardKeyEvent) -> None:
        """Notify the seat that a key has been pressed on the keyboard

        Defers to any keyboard grabs.
        """
        lib.wlr_seat_keyboard_notify_key(
            self._ptr, key_event.time_msec, key_event.keycode, key_event.state
        )

    def keyboard_notify_modifiers(self, modifiers: KeyboardModifiers) -> None:
        """Notify the seat that the modifiers for the keyboard have changed

        Defers to any keyboard grabs.
        """
        lib.wlr_seat_keyboard_notify_modifiers(self._ptr, modifiers._ptr)

    def keyboard_notify_enter(self, surface: Surface, keyboard: Keyboard) -> None:
        """Notify the seat that the keyboard focus has changed

        Notify the seat that the keyboard focus has changed and request it to
        be the focused surface for this keyboard. Defers to any current grab of
        the seat's keyboard.
        """
        lib.wlr_seat_keyboard_notify_enter(
            self._ptr,
            surface._ptr,
            keyboard.keycodes,
            keyboard.num_keycodes,
            keyboard.modifiers._ptr,
        )

    def keyboard_clear_focus(self) -> None:
        """Clear the focused surface for the keyboard and leave all entered surfaces"""
        lib.wlr_seat_keyboard_clear_focus(self._ptr)

    def keyboard_has_grab(self) -> bool:
        """Whether or not the keyboard has a grab other than the default grab"""
        return lib.wlr_seat_keyboard_has_grab(self._ptr)

    def touch_notify_down(
        self,
        surface: Surface,
        time_msec: int,
        touch_id: int,
        surface_x: float,
        surface_y: float,
    ) -> None:
        """
        Notify the seat of a touch down on the given surface. Defers to any grab of the
        touch device.
        """
        lib.wlr_seat_touch_notify_down(
            self._ptr, surface._ptr, time_msec, touch_id, surface_x, surface_y
        )

    def touch_notify_up(self, time_msec: int, touch_id: int) -> None:
        """
        Notify the seat that the touch point given by `touch_id` is up. Defers to any
        grab of the touch device.
        """
        lib.wlr_seat_touch_notify_up(self._ptr, time_msec, touch_id)

    def touch_notify_motion(
        self, time_msec: int, touch_id: int, surface_x: float, surface_y: float
    ) -> None:
        """
        Notify the seat that the touch point given by `touch_id` has moved. Defers to
        any grab of the touch device. The seat should be notified of touch motion even
        if the surface is not the owner of the touch point for processing by grabs.
        """
        lib.wlr_seat_touch_notify_motion(
            self._ptr, time_msec, touch_id, surface_x, surface_y
        )

    def set_selection(self, source, serial: int) -> None:
        """Sets the current selection for the seat

        None can be provided to clear it.  This removes the previous one if
        there was any. In case the selection doesn't come from a client,
        Display.next_serial() can be used to generate a serial.
        """
        if source is None:
            lib.wlr_seat_set_selection(self._ptr, ffi.NULL, serial)
        else:
            # TODO: wrap source in a data source
            lib.wlr_seat_set_selection(self._ptr, source, serial)

    def set_primary_selection(self, source, serial: int) -> None:
        """Sets the current primary selection for the seat.

        None can be provided to clear it. This removes the previous one if
        there was any. In case the selection doesn't come from a client,
        Display.next_serial() can be used to generate a serial.
        """
        if source is None:
            lib.wlr_seat_set_primary_selection(self._ptr, ffi.NULL, serial)
        else:
            # TODO: wrap source in a data source
            lib.wlr_seat_set_primary_selection(self._ptr, source, serial)

    def validate_pointer_grab_serial(self, origin: Surface, serial: int) -> bool:
        """Check whether this serial is valid to start a pointer grab action."""
        return lib.wlr_seat_validate_pointer_grab_serial(self._ptr, origin._ptr, serial)

    def start_pointer_drag(self, drag: Drag, serial: int) -> None:
        """
        Starts a pointer drag on the seat. This starts implicit keyboard and pointer
        grabs.
        """
        lib.wlr_seat_start_pointer_drag(self._ptr, drag._ptr, serial)

    def __enter__(self) -> Seat:
        """Context manager to clean up the seat"""
        return self

    def __exit__(self, exc_type, exc_value, exc_tb) -> None:
        """Clean up the seat when exiting the context"""
        self.destroy()


class PointerRequestSetCursorEvent(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_seat_pointer_request_set_cursor_event *", ptr)

    # TODO: seat client

    @property
    def surface(self) -> Surface:
        # TODO: setup weakref
        return Surface(self._ptr.surface)

    @property
    def serial(self) -> int:
        return self._ptr.serial

    @property
    def hotspot(self) -> tuple[int, int]:
        return self._ptr.hotspot_x, self._ptr.hotspot_y


class RequestSetSelectionEvent(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_seat_request_set_selection_event *", ptr)

    # TODO: source

    @property
    def serial(self) -> int:
        return self._ptr.serial


class RequestSetPrimarySelectionEvent(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast(
            "struct wlr_seat_request_set_primary_selection_event *", ptr
        )

    # TODO: source

    @property
    def serial(self) -> int:
        return self._ptr.serial


class RequestStartDragEvent(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_seat_request_start_drag_event *", ptr)

    @property
    def drag(self) -> Drag:
        return Drag(self._ptr.drag)

    @property
    def origin(self) -> Surface:
        return Surface(self._ptr.origin)

    @property
    def serial(self) -> int:
        return self._ptr.serial


class PointerFocusChangeEvent(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_seat_pointer_focus_change_event *", ptr)

    # TODO


class KeyboardFocusChangeEvent(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_seat_keyboard_focus_change_event *", ptr)

    # TODO


class SeatPointerState(Ptr):
    def __init__(self, ptr) -> None:
        """The current state of the pointer on the seat"""
        self._ptr = ptr

        self.focus_change_event = Signal(
            ptr=ffi.addressof(self._ptr.events.focus_change),
            data_wrapper=PointerFocusChangeEvent,
        )

    @property
    def focused_surface(self) -> Surface | None:
        """The surface that currently has keyboard focus"""
        focused_surface = self._ptr.focused_surface
        if focused_surface == ffi.NULL:
            return None
        return Surface(focused_surface)


class SeatKeyboardState(Ptr):
    def __init__(self, ptr) -> None:
        """The current state of the keyboard on the seat"""
        self._ptr = ptr

        self.focus_change_event = Signal(
            ptr=ffi.addressof(self._ptr.events.focus_change),
            data_wrapper=KeyboardFocusChangeEvent,
        )

    @property
    def focused_surface(self) -> Surface | None:
        """The surface that is currently focused"""
        focused_surface = self._ptr.focused_surface
        if focused_surface == ffi.NULL:
            return None
        return Surface(focused_surface)



# Copyright (c) Sean Vig 2019

from __future__ import annotations

from pywayland.server import Signal

from wlroots import ffi, lib, Ptr
from wlroots.util.box import Box
from .output import Output


class OutputLayout(Ptr):
    def __init__(self) -> None:
        """Creates an output layout to work with a layout of screens

        Creates a output layout, which can be used to describing outputs in
        physical space relative to one another, and perform various useful
        operations on that state.
        """
        ptr = lib.wlr_output_layout_create()
        self._ptr = ffi.gc(ptr, lib.wlr_output_layout_destroy)

        self.add_event = Signal(ptr=ffi.addressof(ptr.events.add))
        self.change_event = Signal(ptr=ffi.addressof(ptr.events.change))
        self.destroy_event = Signal(ptr=ffi.addressof(ptr.events.destroy))

    def destroy(self) -> None:
        """Destroy the current output layout"""
        if self._ptr is not None:
            ffi.release(self._ptr)
            self._ptr = None

    def add_auto(self, output: Output) -> None:
        """Add an auto configured output to the layout

        This will place the output in a sensible location in the layout. The
        coordinates of the output in the layout may adjust dynamically when the
        layout changes. If the output is already in the layout, it will become
        auto configured. If the position of the output is set such as with
        `wlr_output_layout_move()`, the output will become manually configured.

        :param output:
            The output to configure the layout against.
        """
        lib.wlr_output_layout_add_auto(self._ptr, output._ptr)

    def output_coords(self, output: Output) -> tuple[float, float]:
        """Determine coordinates of the output in the layout

        Given x and y in layout coordinates, adjusts them to local output
        coordinates relative to the given reference output.
        """
        ox = ffi.new("double *")
        oy = ffi.new("double *")
        lib.wlr_output_layout_output_coords(self._ptr, output._ptr, ox, oy)

        return ox[0], oy[0]

    def __enter__(self) -> OutputLayout:
        """Use the output layout in a context manager"""
        return self

    def __exit__(self, exc_type, exc_value, exc_tb) -> None:
        """Clean up the output layout when exiting the context"""
        self.destroy()

    def output_at(self, x: float, y: float) -> Output | None:
        """
        Get the output at the specified layout coordinates. Returns None if no output
        matches the coordinates.
        """
        output_ptr = lib.wlr_output_layout_output_at(self._ptr, x, y)
        if output_ptr == ffi.NULL:
            return None
        return Output(output_ptr)

    def add(self, output: Output, lx: int, ly: int) -> None:
        """
        Add the output to the layout at the specified coordinates. If the output is
        already part of the output layout, this moves the output.
        """
        lib.wlr_output_layout_add(self._ptr, output._ptr, lx, ly)

    def move(self, output: Output, lx: int, ly: int) -> None:
        """Move an output to specified coordinates."""
        lib.wlr_output_layout_move(self._ptr, output._ptr, lx, ly)

    def remove(self, output: Output) -> None:
        """Remove an output from the layout."""
        lib.wlr_output_layout_remove(self._ptr, output._ptr)

    def get_box(self, reference: Output | None = None) -> Box:
        """
        Get the box of the layout for the given reference output in layout
        coordinates. If `reference` is None, the box will be for the extents of the
        entire layout.
        """
        if reference:
            box_ptr = lib.wlr_output_layout_get_box(self._ptr, reference._ptr)
        else:
            box_ptr = lib.wlr_output_layout_get_box(self._ptr, ffi.NULL)
        return Box(ptr=box_ptr)

    def closest_point(
        self, lx: float, ly: float, reference: Output | None = None
    ) -> tuple[float, float]:
        """
        Get the closest point on this layout from the given point from the reference
        output. If reference is NULL, gets the closest point from the entire layout.
        """
        if reference:
            reference_ptr = reference._ptr
        else:
            reference_ptr = ffi.NULL

        dest_lx = ffi.new("double *")
        dest_ly = ffi.new("double *")
        lib.wlr_output_layout_closest_point(
            self._ptr, reference_ptr, lx, ly, dest_lx, dest_ly
        )
        return dest_lx[0], dest_ly[0]



# Copyright (c) Sean Vig 2022

from __future__ import annotations

import weakref

from wlroots import ffi, lib, PtrHasData, Ptr
from wlroots.util.clock import Timespec
from wlroots.wlr_types import Output, OutputLayout
from wlroots.wlr_types.xdg_shell import XdgSurface

_weakkeydict: weakref.WeakKeyDictionary = weakref.WeakKeyDictionary()


class Scene(Ptr):
    def __init__(self, output_layout: OutputLayout) -> None:
        """ "A root scene-graph node."""
        self._ptr = lib.wlr_scene_create()
        if not lib.wlr_scene_attach_output_layout(self._ptr, output_layout._ptr):
            raise RuntimeError("Unable to attach scene to output layout")

    @property
    def node(self) -> SceneNode:
        """The associated scene node."""
        node = SceneNode(ffi.addressof(self._ptr.node))
        _weakkeydict[node] = self._ptr
        return node

    def get_scene_output(self, output: Output) -> SceneOutput:
        """Get a scene-graph output from a wlr_output."""
        ptr = lib.wlr_scene_get_scene_output(self._ptr, output._ptr)
        return SceneOutput(ptr)


class SceneOutput(Ptr):
    def __init__(self, ptr) -> None:
        """A viewport for an output in the scene-graph"""
        self._ptr = ptr

    def commit(self) -> None:
        """Render and commit an output."""
        if not lib.wlr_scene_output_commit(self._ptr):
            raise RuntimeError("Unable to commit scene output")

    def send_frame_done(self, timespec: Timespec) -> None:
        """Send frame done on scene output.

        Call send_frame_done() on all surfaces in the scene rendered by
        commit() for which scene_surface.primary_output matches the given
        scene_output.
        """
        lib.wlr_scene_output_send_frame_done(self._ptr, timespec._ptr)


class SceneNode(PtrHasData):
    def __init__(self, ptr) -> None:
        """ "A node is an object in the scene."""
        self._ptr = ptr

    @classmethod
    def xdg_surface_create(
        cls, parent: SceneNode, xdg_surface: XdgSurface
    ) -> SceneNode:
        """Add a node displaying an xdg_surface and all of its sub-surfaces to the scene-graph.

        The origin of the returned scene-graph node will match the top-left
        corner of the xdg_surface window geometry.
        """
        ptr = lib.wlr_scene_xdg_surface_create(parent._ptr, xdg_surface._ptr)
        return SceneNode(ptr)

    def set_position(self, x: int, y: int) -> None:
        """Set the output's position in the scene-graph."""
        lib.wlr_scene_node_set_position(self._ptr, x, y)

    def raise_to_top(self) -> None:
        """Move the node above all of its sibling nodes."""
        lib.wlr_scene_node_raise_to_top(self._ptr)

    def lower_to_bottom(self) -> None:
        """Move the node below all of its sibling nodes."""
        lib.wlr_scene_node_lower_to_bottom(self._ptr)

    def place_above(self, sibling: SceneNode) -> None:
        """Move the node right above the specified sibling."""
        lib.wlr_scene_node_place_above(self._ptr, sibling._ptr)

    def place_below(self, sibling: SceneNode) -> None:
        """Move the node right below the specified sibling."""
        lib.wlr_scene_node_place_below(self._ptr, sibling._ptr)



# Copyright (c) Matt Colligan 2021

from pywayland.server import Display

from wlroots import lib, Ptr


class PrimarySelectionV1DeviceManager(Ptr):
    def __init__(self, display: Display) -> None:
        """Data manager to handle the primary selection

        :param display:
            The display to handle the clipboard for
        """
        self._ptr = lib.wlr_primary_selection_v1_device_manager_create(display._ptr)



"""An example basic compositor

To run this example, install the wlroots required dependencies, build the ffi
bindings (`python wlroots/ffi_build.py`), and launch the main (`python -m tiny`).
"""
from __future__ import annotations

import logging
import sys

from pywayland.server import Display
from wlroots.helper import build_compositor
from wlroots.util.log import log_init
from wlroots.wlr_types import (
    Cursor,
    DataDeviceManager,
    OutputLayout,
    Scene,
    Seat,
    XCursorManager,
    XdgShell,
)

from .server import TinywlServer


def main(argv) -> None:
    with Display() as display:
        compositor, allocator, renderer, backend = build_compositor(display)
        device_manager = DataDeviceManager(display)  # noqa: F841
        xdg_shell = XdgShell(display)
        with OutputLayout() as output_layout, Cursor(
            output_layout
        ) as cursor, XCursorManager(24) as xcursor_manager, Seat(
            display, "seat0"
        ) as seat:
            scene = Scene(output_layout)
            tinywl_server = TinywlServer(  # noqa: F841
                display=display,
                backend=backend,
                allocator=allocator,
                renderer=renderer,
                scene=scene,
                xdg_shell=xdg_shell,
                cursor=cursor,
                cursor_manager=xcursor_manager,
                seat=seat,
                output_layout=output_layout,
            )

            socket = display.add_socket()
            print("socket:", socket.decode())
            with backend:
                display.run()


if __name__ == "__main__":
    log_init(logging.DEBUG)
    logging.basicConfig(level=logging.INFO)

    main(sys.argv[1:])



from pywayland.server import Display, Signal

from wlroots import Ptr, ffi, lib

from .seat import Seat


class IdleTimeout(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_idle_timeout *", ptr)

        self.idle_event = Signal(ptr=ffi.addressof(self._ptr.events.idle))
        self.resume_event = Signal(ptr=ffi.addressof(self._ptr.events.resume))
        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))

    @property
    def idle_state(self) -> bool:
        return self._ptr.idle_state

    @property
    def enabled(self) -> bool:
        return self._ptr.enabled

    @property
    def timeout(self) -> int:
        return self._ptr.timeout

    def destroy(self) -> None:
        if self._ptr is not None:
            ffi.release(self._ptr)
            self._ptr = None


class Idle(Ptr):
    def __init__(self, display: Display) -> None:
        self._ptr = lib.wlr_idle_create(display._ptr)

        self.activity_notify_event = Signal(
            ptr=ffi.addressof(self._ptr.events.activity_notify)
        )
        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))

    @property
    def enabled(self) -> bool:
        return self._ptr.enabled

    def notify_activity(self, seat: Seat) -> None:
        """
        Send notification to restart all timers for the given seat. Called by
        compositor when there is an user activity event on that seat.
        """
        lib.wlr_idle_notify_activity(self._ptr, seat._ptr)

    def set_enabled(self, seat: Seat, enabled: bool):
        """
        Enable or disable timers for a given idle resource by seat.
        Passing a NULL seat means update timers for all seats.
        """
        lib.wlr_idle_set_enabled(self._ptr, seat._ptr, enabled)

    def idle_timeout_create(self, seat: Seat, timeout: int) -> IdleTimeout:
        """
        Create a new timer on the given seat. The idle event will be called after
        the given amount of milliseconds of inactivity, and the resumed event will
        be sent at the first user activity after the fired event.
        """
        return IdleTimeout(lib.wlr_idle_timeout_create(self._ptr, seat._ptr, timeout))



# Copyright (c) Matt Colligan 2021

from __future__ import annotations

import enum
from typing import TYPE_CHECKING
from weakref import WeakKeyDictionary

from pywayland.server import Signal

from wlroots import ffi, lib, Ptr
from .surface import Surface
from wlroots.util.region import PixmanRegion32

if TYPE_CHECKING:
    from pywayland.server import Display

_weakkeydict: WeakKeyDictionary = WeakKeyDictionary()


class PointerConstraintV1Type(enum.IntEnum):
    LOCKED = lib.WLR_POINTER_CONSTRAINT_V1_LOCKED
    CONFINED = lib.WLR_POINTER_CONSTRAINT_V1_CONFINED


class PointerConstraintV1StateField(enum.IntEnum):
    REGION = lib.WLR_POINTER_CONSTRAINT_V1_STATE_REGION
    CURSOR_HINT = lib.WLR_POINTER_CONSTRAINT_V1_STATE_CURSOR_HINT


class PointerConstraintsV1(Ptr):
    def __init__(self, display: Display) -> None:
        """Manager to handle pointer constraint requests.

        :param display:
            The display to manage pointer constraints on.
        """
        self._ptr = lib.wlr_pointer_constraints_v1_create(display._ptr)

        self.new_constraint_event = Signal(
            ptr=ffi.addressof(self._ptr.events.new_constraint),
            data_wrapper=PointerConstraintV1,
        )


class PointerConstraintV1(Ptr):
    def __init__(self, ptr) -> None:
        """A `struct wlr_pointer_constraint_v1` instance."""
        self._ptr = ffi.cast("struct wlr_pointer_constraint_v1 *", ptr)

        self.set_region_event = Signal(
            ptr=ffi.addressof(self._ptr.events.set_region),
        )
        self.destroy_event = Signal(
            ptr=ffi.addressof(self._ptr.events.destroy),
            data_wrapper=PointerConstraintV1,
        )

    def send_activated(self) -> None:
        lib.wlr_pointer_constraint_v1_send_activated(self._ptr)

    def send_deactivated(self) -> None:
        lib.wlr_pointer_constraint_v1_send_deactivated(self._ptr)

    @property
    def surface(self) -> Surface:
        surface_ptr = self._ptr.surface
        _weakkeydict[surface_ptr] = self._ptr
        return Surface(surface_ptr)

    @property
    def type(self) -> PointerConstraintV1Type:
        return PointerConstraintV1Type(self._ptr.type)

    @property
    def region(self) -> PixmanRegion32:
        region_ptr = ffi.addressof(self._ptr, "region")
        return PixmanRegion32(region_ptr)

    @property
    def current(self) -> PointerConstraintV1State:
        return PointerConstraintV1State(self._ptr.current)

    @property
    def pending(self) -> PointerConstraintV1State:
        return PointerConstraintV1State(self._ptr.pending)


class PointerConstraintV1State(Ptr):
    def __init__(self, ptr):
        self._ptr = ptr

    @property
    def committed(self) -> PointerConstraintV1StateField:
        return PointerConstraintV1StateField(self._ptr.committed)

    @property
    def region(self) -> PixmanRegion32:
        return PixmanRegion32(self._ptr.region)

    @property
    def cursor_hint(self) -> tuple[float, float]:
        return self._ptr.cursor_hint.x, self._ptr.cursor_hint.y



from pywayland.server import Display

from wlroots.backend import Backend, BackendType


def test_run_headless_env_var(headless_backend):
    with Display() as display:
        with Backend(display):
            pass


def test_run_headless_arg(headless_backend):
    with Display() as display:
        with Backend(display, backend_type=BackendType.HEADLESS):
            pass



# Copyright (c) Sean Vig 2019
from __future__ import annotations

import enum
import weakref

from .keyboard import Keyboard
from wlroots import ffi, PtrHasData, lib

_weakkeydict: weakref.WeakKeyDictionary = weakref.WeakKeyDictionary()


@enum.unique
class ButtonState(enum.IntEnum):
    RELEASED = lib.WLR_BUTTON_RELEASED
    PRESSED = lib.WLR_BUTTON_PRESSED


@enum.unique
class InputDeviceType(enum.IntEnum):
    KEYBOARD = lib.WLR_INPUT_DEVICE_KEYBOARD
    POINTER = lib.WLR_INPUT_DEVICE_POINTER
    TOUCH = lib.WLR_INPUT_DEVICE_TOUCH
    TABLET_TOOL = lib.WLR_INPUT_DEVICE_TABLET_TOOL
    TABLET_PAD = lib.WLR_INPUT_DEVICE_TABLET_PAD
    SWITCH = lib.WLR_INPUT_DEVICE_SWITCH


class InputDevice(PtrHasData):
    def __init__(self, ptr) -> None:
        """Create the input device from the given cdata

        :param ptr:
            The wlr_input_device for the given input device
        """
        self._ptr = ffi.cast("struct wlr_input_device *", ptr)

    @property
    def device_type(self) -> InputDeviceType:
        """The device type associated with the current device"""
        return InputDeviceType(self._ptr.type)

    @property
    def vendor(self) -> int:
        return self._ptr.vendor

    @property
    def product(self) -> int:
        return self._ptr.product

    @property
    def name(self) -> str:
        return ffi.string(self._ptr.name).decode()

    @property
    def keyboard(self) -> Keyboard:
        """Return the keyboard type associated with the input device

        The device must be a keyboard, otherwise this throws a ValueError.
        """
        if self.device_type != InputDeviceType.KEYBOARD:
            raise ValueError(
                f"Expected keyborad input device type, got: {self.device_type}"
            )

        keyboard = Keyboard(self._ptr.keyboard)

        _weakkeydict[keyboard] = self._ptr

        return keyboard

    def libinput_get_device_handle(self) -> ffi.CData | None:
        """
        Returns the underlying libinput device if there is one.

        Returns a pointer to a `struct libinput_device` for use by libinput-interfacing
        code.
        """
        if lib.wlr_input_device_is_libinput(self._ptr):
            return lib.wlr_libinput_get_device_handle(self._ptr)
        return None



from __future__ import annotations

import argparse
import pathlib
import subprocess
import sys
import tempfile

INCLUDE_PATH = pathlib.Path(__file__).parent / "wlroots" / "include"
WAYLAND_PROCOTOLS = [
    "stable/xdg-shell/xdg-shell.xml",
    "unstable/idle-inhibit/idle-inhibit-unstable-v1.xml",
    "unstable/pointer-constraints/pointer-constraints-unstable-v1.xml",
]
WLROOTS_PROTOCOLS = [
    "protocol/idle.xml",
    "protocol/wlr-output-power-management-unstable-v1.xml",
    "protocol/wlr-layer-shell-unstable-v1.xml",
]


def get_wayland_protocols_dir() -> pathlib.Path | None:
    args = ["pkgconf", "--variable=pkgdatadir", "wayland-protocols"]
    try:
        output = subprocess.check_output(args).decode().strip()
    except subprocess.CalledProcessError:
        return None

    if output:
        return pathlib.Path(output)
    return None


def header_filename(xml_file: pathlib.Path) -> str:
    return f"{xml_file.stem}-protocol.h"


def generate_protocol_header(
    input_xml: pathlib.Path, output_dir: pathlib.Path
) -> pathlib.Path:
    output_path = output_dir / header_filename(input_xml)
    subprocess.check_output(
        ["wayland-scanner", "server-header", str(input_xml), str(output_path)]
    )
    return output_path


def check(protocols: list[pathlib.Path]) -> None:
    expected_protocol_files = {
        header_filename(protocol_xml) for protocol_xml in protocols
    }
    protocol_files = {path.name for path in INCLUDE_PATH.iterdir()}

    if expected_protocol_files != protocol_files:
        unexpected_files = list(expected_protocol_files - protocol_files)
        extra_files = list(protocol_files - expected_protocol_files)
        raise ValueError(
            f"Unexpected protocol files: {unexpected_files}, Extra files: {extra_files}"
        )

    for protocol_xml in protocols:
        protocol_header_file = INCLUDE_PATH / header_filename(protocol_xml)
        with protocol_header_file.open() as f:
            protocol_header = f.readlines()

        with tempfile.TemporaryDirectory() as output_dir:
            generated_file = generate_protocol_header(
                protocol_xml, pathlib.Path(output_dir)
            )
            with generated_file.open() as f:
                generated_header = f.readlines()

        if len(protocol_header) != len(generated_header):
            raise ValueError(f"Mismatch line count in {protocol_xml.name}")

        for i, (generated_line, line) in enumerate(
            zip(generated_header, protocol_header)
        ):
            if generated_line != line:
                print("Expected:", generated_line.strip("\n"))
                print("Got:", line.strip("\n"))
                raise ValueError(f"Mismatch in {protocol_xml.name} in line {i}")


def generate(protocols: list[pathlib.Path]) -> None:
    for protocol_xml in protocols:
        generate_protocol_header(protocol_xml, INCLUDE_PATH)


def parse_args(argv):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--wayland-dir", default=get_wayland_protocols_dir(), type=pathlib.Path
    )
    parser.add_argument("--wlroots-dir", required=True, type=pathlib.Path)
    parser.add_argument("--generate", action="store_true")
    args = parser.parse_args(argv)

    if args.wayland_dir is None or not args.wayland_dir.exists():
        raise ValueError(
            f"Wayland protocols directory does not exist: {args.wayland_dir}"
        )
    if not args.wlroots_dir.exists():
        raise ValueError(f"Wlroots directory does not exist: {args.wlroots_dir}")

    protocols = [args.wayland_dir / protocol for protocol in WAYLAND_PROCOTOLS] + [
        args.wlroots_dir / protocol for protocol in WLROOTS_PROTOCOLS
    ]

    for protocol in protocols:
        if not protocol.exists():
            raise ValueError(f"Protocol does not exist: {protocol}")

    return protocols, args.generate


if __name__ == "__main__":
    protocols, generate_headers = parse_args(sys.argv[1:])

    if generate_headers:
        generate(protocols)
    else:
        check(protocols)



# Copyright (c) Matt Colligan 2021

from wlroots import ffi, Ptr
from .input_device import InputDevice


class TouchEventUp(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_event_touch_up *", ptr)

    @property
    def device(self) -> InputDevice:
        return InputDevice(self._ptr.device)

    @property
    def time_msec(self) -> int:
        return self._ptr.time_msec

    @property
    def touch_id(self) -> int:
        return self._ptr.touch_id


class TouchEventDown(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_event_touch_down *", ptr)

    @property
    def device(self) -> InputDevice:
        return InputDevice(self._ptr.device)

    @property
    def time_msec(self) -> int:
        return self._ptr.time_msec

    @property
    def touch_id(self) -> int:
        return self._ptr.touch_id

    @property
    def x(self) -> float:
        return self._ptr.x

    @property
    def y(self) -> float:
        return self._ptr.y


class TouchEventMotion(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_event_touch_motion *", ptr)

    @property
    def device(self) -> InputDevice:
        return InputDevice(self._ptr.device)

    @property
    def time_msec(self) -> int:
        return self._ptr.time_msec

    @property
    def touch_id(self) -> int:
        return self._ptr.touch_id

    @property
    def x(self) -> float:
        return self._ptr.x

    @property
    def y(self) -> float:
        return self._ptr.y


class TouchEventCancel(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_event_touch_cancel *", ptr)

    @property
    def device(self) -> InputDevice:
        return InputDevice(self._ptr.device)

    @property
    def time_msec(self) -> int:
        return self._ptr.time_msec

    @property
    def touch_id(self) -> int:
        return self._ptr.touch_id
